import os
from typing import Union, List, Callable

import numpy as np
import matplotlib.pyplot as plt

from ._explain import ExplainerMixin


# todo, optionally show predicted value as dots on plots

def compute_bounds(xmin, xmax, xv):
    """
    from shap
    Handles any setting of xmax and xmin.

    Note that we handle None, float, or "percentile(float)" formats.
    """

    if xmin is not None or xmax is not None:
        if type(xmin) == str and xmin.startswith("percentile"):
            xmin = np.nanpercentile(xv, float(xmin[11:-1]))
        if type(xmax) == str and xmax.startswith("percentile"):
            xmax = np.nanpercentile(xv, float(xmax[11:-1]))

        if xmin is None or xmin == np.nanmin(xv):
            xmin = np.nanmin(xv) - (xmax - np.nanmin(xv))/20
        if xmax is None or xmax == np.nanmax(xv):
            xmax = np.nanmax(xv) + (np.nanmax(xv) - xmin)/20

    return xmin, xmax


class PartialDependencePlot(ExplainerMixin):
    """
    Partial dependence plots as introduced by Friedman_ et al., 2001

    Example
    -------
        >>> from ai4water import Model
        >>> from ai4water.datasets import busan_beach
        >>> from ai4water.postprocessing.explain import PartialDependencePlot
        >>> data = busan_beach()
        >>> model = Model(model="XGBRegressor")
        >>> model.fit(data=data)
        # get the data to explain
        >>> x, _ = model.training_data()
        >>> pdp = PartialDependencePlot(model.predict, x, model.input_features,
        >>>                            num_points=14)

    .. _Friedman:
        https://doi.org/10.1214/aos/1013203451
    """
    def __init__(
            self,
            model: Callable,
            data,
            feature_names=None,
            num_points: int = 100,
            path=None,
            **kwargs
    ):
        """Initiates the class

        Parameters
        ----------
            model :
                the trained/calibrated model which must be callable. It must take the
                `data` as input and sprout an array of predicted values. For example
                if you are using Keras/sklearn model, then you must pass model.predict
            data :
                The inputs to the `model`.
            feature_names :
                Names of features. Used for labeling.
            num_points :
                determines the grid for evaluation of `model`
            path:
                path to save the plots. By default the results are saved in current directory
            kwargs:
                any additional keyword arguments for `model`
        """

        self.model = model
        self.num_points = num_points
        self.xmin = "percentile(0)"
        self.xmax = "percentile(100)"
        self.kwargs = kwargs

        super().__init__(data=data, features=feature_names, path=path or os.getcwd())

    def nd_interactions(
            self,
            height: int = 2,
            ice: bool = False,
            show_dist: bool = False,
            show_minima: bool = False,
            show: bool = True,
            save: bool = False
    ) -> plt.Figure:
        """Plots 2d interaction plots of all features as done in skopt

        Arguments:
            height:
                height of each subplot in inches
            ice:
                whether to show the ice lines or not
            show_dist:
                whether to show the distribution of data as histogram or not
            show_minima:
                whether to show the function minima or not
            show:
                whether to show the plot or not
            save:
                whether to save the plot or not
        Returns:
            matplotlib Figure
        """

        n_dims = len(self.features)

        fig, ax = plt.subplots(n_dims, n_dims, figsize=(height * n_dims, height * n_dims))

        fig.subplots_adjust(left=0.05, right=0.95, bottom=0.05, top=0.95,
                            hspace=0.1, wspace=0.1)

        for i in range(n_dims):
            for j in range(n_dims):
                # diagonal
                if i == j:

                    if n_dims > 1:
                        ax_ = ax[i, i]
                    else:
                        ax_ = ax

                    self._plot_pdp_1dim(*self._pdp_for_2d(self.data, self.features[i]),
                                        self.data, self.features[i],
                                        show_dist=show_dist, show_minima=show_minima,
                                        ice=ice, show=False, ax=ax_)
                    # resetting the label
                    # ax_.set_xlabel(self.features[i])
                    ax_.set_ylabel(self.features[i])
                    process_axis(ax_, xlabel=self.features[i], top_spine=True, right_spine=True)

                # lower triangle
                elif i > j:
                    self.plot_interaction(
                        features=[self.features[j], self.features[i]],
                        show=False, ax=ax[i, j], colorbar=False
                    )

                elif j > i:
                    if not ax[i, j].lines:  # empty axes
                        ax[i, j].axis("off")

                if j > 0:  # not the left most column
                    ax[i, j].yaxis.set_ticks([])
                    ax[i, j].yaxis.set_visible(False)
                    ax[i, j].yaxis.label.set_visible(False)

                if i < n_dims-1:  # not the bottom most row
                    ax[i, j].xaxis.set_ticks([])
                    ax[i, j].xaxis.set_visible(False)
                    ax[i, j].xaxis.label.set_visible(False)

        if save:
            fname = os.path.join(self.path, f"pdp_interact_nd")
            plt.savefig(fname, bbox_inches="tight", dpi=100*n_dims)

        if show:
            plt.show()

        return fig

    def plot_interaction(
            self,
            features: list,
            lookback: int = None,
            ax: plt.Axes = None,
            plot_type: str = "2d",
            cmap=None,
            colorbar: bool = True,
            show: bool = True,
            save: bool = False,
            **kwargs
    ) -> plt.Axes:
        """Shows interaction between two features

        Arguments:
            features:
                a list or tuple of two feature names to use
            lookback:

            ax:
                matplotlib axes on which to draw. If not given, current axes will
                be used.
            plot_type:
                either "2d" or "surface"
            cmap:
                color map to use
            colorbar:
                whether to show the colorbar or not
            show:
                whether to show the plot or not
            save:
                whether to save the plot or not
            kwargs:
                any keyword argument for axes.plot_surface or axes.contourf

        Returns:
              matplotlib Axes

        Examples
        --------
        >>> from ai4water import Model
        >>> from ai4water.datasets import busan_beach
        >>> from ai4water.postprocessing.explain import PartialDependencePlot
        >>> data = busan_beach()
        >>> model = Model(model="XGBRegressor")
        >>> model.fit(data=busan_beach())
        >>> x, _ = model.training_data()
        >>> pdp = PartialDependencePlot(model.predict, x, model.input_features,
        ...                            num_points=14)
        >>> pdp.nd_interactions(show_dist=True)

        specifying features whose interaction is to be calculated and plotted.

        >>> axis = pdp.plot_interaction(["tide_cm", "wat_temp_c"])
        """

        if not self.data_is_2d:
            raise NotImplementedError

        assert isinstance(features, list) and len(features) == 2

        x0, x1, pd_vals = self._interaction(features, self.data, lookback)

        kwds = {}
        if plot_type == "surface":
            kwds['projection'] = '3d'

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, **kwds)

        if plot_type == "surface":
            _add_surface(ax, x0, x1, pd_vals, cmap, features[0], features[1], **kwargs)
        else:
            self._plot_interaction(ax, x0, x1, pd_vals,
                                   cmap=cmap,
                                   features=features,
                                   lookback=lookback,
                                   colorbar=colorbar, **kwargs)

        if save:
            fname = os.path.join(self.path, f"pdp_interact{features[0]}_{features[1]}")
            plt.savefig(fname, bbox_inches="tight", dpi=300)

        if show:
            plt.show()

        return ax

    def _plot_interaction(self, ax, x0, x1, pd_vals, cmap, features, lookback,
                          colorbar=True,
                          **kwargs):
        """adds a 2d interaction plot"""
        cntr = ax.contourf(x0, x1, pd_vals, cmap=cmap, **kwargs)

        xv0 = self.xv(self.data, features[0], lookback)
        xv1 = self.xv(self.data, features[1], lookback)

        ax.scatter(xv0, xv1, c='k', s=10, lw=0.)
        process_axis(ax, xlabel=features[0], ylabel=features[1])

        if colorbar:
            cbar = plt.colorbar(cntr, ax=ax)
            cbar.set_label(f"E[f(x) | {features[0]} {features[1]} ]", rotation=90)

        return

    def _interaction(self, features, data, lookback):

        ind0 = self._feature_to_ind(features[0])
        ind1 = self._feature_to_ind(features[1])

        xs0 = self.grid(self.data, features[0], lookback)
        xs1 = self.grid(self.data, features[1], lookback)

        features_tmp = data.copy()
        x0 = np.zeros((self.num_points, self.num_points))
        x1 = np.zeros((self.num_points, self.num_points))
        pd_vals = np.zeros((self.num_points, self.num_points))

        for i in range(self.num_points):
            for j in range(self.num_points):
                features_tmp[:, ind0] = xs0[i]
                features_tmp[:, ind1] = xs1[j]
                x0[i, j] = xs0[i]
                x1[i, j] = xs1[j]
                pd_vals[i, j] = self.model(features_tmp).mean()

        return x0, x1, pd_vals

    def plot_1d(
            self,
            feature,
            show_dist: bool = True,
            show_dist_as: str = "hist",
            ice: bool = True,
            feature_expected_value: bool = False,
            model_expected_value: bool = False,
            show_ci: bool = False,
            show_minima: bool = False,
            show: bool = True,
            save: bool = False
    ):
        """partial dependence plot in one dimension

        Arguments:
            feature:
                the feature name for which to plot the partial dependence
            show_dist:
                whether to show actual distribution of data or not
            show_dist_as:
                one of "hist" or "grid"
            ice:
                whether to show individual component elements on plot or not
            feature_expected_value:
                whether to show the average value of feature on the plot or not
            model_expected_value:
                whether to show average prediction on plot or not
            show_ci:
                whether to show confidence interval of pdp or not
            show_minima:
                whether to indicate the minima or not
            show:
                whether to show the plot or not
            save:
                whether to save the plot or not
        """
        if isinstance(feature, list) or isinstance(feature, tuple):
            raise NotImplementedError
        else:
            if self.single_source:
                if self.data_is_2d:
                    self._plot_pdp_1dim(*self._pdp_for_2d(self.data, feature), self.data, feature,
                                        show_dist=show_dist, show_dist_as=show_dist_as,
                                        ice=ice, feature_expected_value=feature_expected_value,
                                        show_ci=show_ci, show_minima=show_minima,
                                        model_expected_value=model_expected_value, show=show, save=save)
                elif self.data_is_3d:
                    for lb in range(self.data.shape[1]):
                        self._plot_pdp_1dim(*self._pdp_for_2d(self.data, feature, lb),
                                            data=self.data, feature=feature, lookback=lb,
                                            show_ci=show_ci, show_minima=show_minima,
                                            show_dist=show_dist, show_dist_as=show_dist_as,
                                            ice=ice, feature_expected_value=feature_expected_value,
                                            model_expected_value=model_expected_value, show=show, save=save)
            else:
                for data in self.data:
                    if self.data_is_2d:
                        self._pdp_for_2d(data, feature)
                    else:
                        for lb in []:
                            self._pdp_for_2d(data, feature, lb)
        return

    def xv(self, data, feature, lookback):

        ind = self._feature_to_ind(feature)
        if data.ndim == 3:
            xv = data[:, lookback, ind]
        else:
            xv = data[:, ind]

        return xv

    def grid(self, data, feature, lookback=None):
        """generates the grid for evaluation of model"""
        xmin, xmax = compute_bounds(self.xmin, self.xmax, self.xv(data, feature, lookback))
        return np.linspace(xmin, xmax, self.num_points)

    def _pdp_for_2d(self, data, feature, lookback=None):
        ind = self._feature_to_ind(feature)

        xs = self.grid(data, feature, lookback)

        data_temp = data.copy()
        pd_vals = np.full(self.num_points, np.nan)
        ice_vals = np.full((self.num_points, data.shape[0]), np.nan)

        for i in range(self.num_points):

            if data.ndim == 3:
                data_temp[:, lookback, ind] = xs[i]
            else:
                data_temp[:, ind] = xs[i]

            preds = self.model(data_temp, **self.kwargs)
            pd_vals[i] = preds.mean()
            ice_vals[i, :] = preds.reshape(-1,)

        return pd_vals, ice_vals

    def _feature_to_ind(self, feature) -> int:
        ind = feature
        if isinstance(feature, str):
            if self.single_source:
                ind = self.features.index(feature)
            else:
                raise NotImplementedError
        elif not isinstance(feature, int):
            raise ValueError

        return ind

    def _plot_pdp_1dim(self,
                       pd_vals, ice_vals, data, feature, lookback=None,
                       show_dist=True, show_dist_as="hist", ice=True, show_ci=False,
                       show_minima=False,
                       feature_expected_value=False, model_expected_value=False,
                       show=True, save=False, ax=None):

        xmin, xmax = compute_bounds(self.xmin, self.xmax, self.xv(data, feature, lookback))

        if ax is None:
            fig = plt.figure()
            ax = fig.add_axes((0.1, 0.3, 0.8, 0.6))

        xs = self.grid(data, feature, lookback)

        ylabel = "E[f(x) | " + feature + "]"
        if ice:
            ice_linewidth = min(1, 50 / ice_vals.shape[1])  # pylint: disable=unsubscriptable-object
            ax.plot(xs, ice_vals, color='lightblue', linewidth=ice_linewidth, alpha=1)
            ylabel = "f(x) | " + feature

        if show_ci:
            std = np.std(ice_vals, axis=1)
            upper = pd_vals + std
            lower = pd_vals - std
            ax.fill_between(xs, upper, lower, alpha=0.14, color='#66C2D7')

        # the line plot
        ax.plot(xs, pd_vals, color='blue', linewidth=2, alpha=1)

        title = None
        if lookback is not None:
            title = f"lookback: {lookback}"
        process_axis(ax, ylabel=ylabel, ylabel_kws=dict(fontsize=20), right_spine=False, top_spine=False,
                     tick_params=dict(labelsize=11), xlabel=feature, xlabel_kws=dict(fontsize=20), title=title)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')

        ax2 = ax.twinx()

        if show_dist:
            xv = self.xv(data, feature, lookback)
            if show_dist_as == "hist":

                ax2.hist(xv, 50, density=False, facecolor='black', alpha=0.1, range=(xmin, xmax))
            else:
                _add_dist_as_grid(fig, xv, other_axes=ax, xlabel=feature, xlabel_kws=dict(fontsize=20))

        process_axis(ax2, right_spine=False, top_spine=False, left_spine=False, bottom_spine=False,
                     ylim=(0, data.shape[0]))
        ax2.xaxis.set_ticks_position('bottom')
        ax2.yaxis.set_ticks_position('left')
        ax2.yaxis.set_ticks([])

        if feature_expected_value:
            self._add_feature_exp_val(ax2, ax, xmin, xmax, data, feature, lookback)

        if model_expected_value:
            self._add_model_exp_val(ax2, ax, data)

        if show_minima:
            minina = self.model(data, **self.kwargs).min()
            ax.axvline(minina, linestyle="--", color="r", lw=1)

        if save:
            lookback = lookback or ''
            fname = os.path.join(self.path, f"pdp_{feature}_{lookback}")
            plt.savefig(fname, bbox_inches="tight", dpi=400)

        if show:
            plt.show()

        return ax

    def _add_model_exp_val(self, ax, original_axis, data):
        """adds model expected value on a duplicate axis of ax"""
        model_expected_val = self.model(data, **self.kwargs).mean()

        ax2 = ax.twinx()

        ymin, ymax = original_axis.get_ylim()
        process_axis(ax2, ylim=(ymin, ymax),
                     yticks=[model_expected_val],
                     yticklabels=["E[f(x)]"],
                     right_spine=False,
                     top_spine=False,
                     tick_params=dict(length=0, labelsize=11)
                     )

        original_axis.axhline(model_expected_val, color="#999999", zorder=-1, linestyle="--", linewidth=1)
        return

    def _add_feature_exp_val(self, ax, original_axis, xmin, xmax, data, feature, lookback=None):

        xv = self.xv(data=data, feature=feature, lookback=lookback)
        mval = xv.mean()

        ax3 = ax.twiny()

        process_axis(ax3,
                     xlim=(xmin, xmax),
                     xticks=[mval], xticklabels=["E[" + feature + "]"],
                     tick_params={'length': 0, 'labelsize': 11}, top_spine=False, right_spine=False)
        original_axis.axvline(mval, color="#999999", zorder=-1, linestyle="--", linewidth=1)
        return


def process_axis(
        ax: plt.Axes,
        title=None, title_kws=None,
        xlabel=None, xlabel_kws=None,
        ylabel=None, ylabel_kws=None,
        xticks=None, xticklabels=None,
        yticks=None, yticklabels=None,
        tick_params=None,
        top_spine=None, right_spine=None, bottom_spine=None, left_spine=None,
        xlim=None, ylim=None
):
    """processes a matplotlib axes"""
    if title:
        title_kws = title_kws or {}
        ax.set_title(title, **title_kws)

    if ylabel:
        ylabel_kws = ylabel_kws or {}
        ax.set_ylabel(ylabel, **ylabel_kws)

    if xlabel:
        xlabel_kws = xlabel_kws or {}
        ax.set_xlabel(xlabel, **xlabel_kws)

    if xlim is not None:
        ax.set_xlim(xlim)

    if ylim is not None:
        ax.set_ylim(ylim)

    if xticks:
        ax.set_xticks(xticks)

    if xticklabels is not None:
        ax.set_xticklabels(xticklabels)

    if yticks:
        ax.set_yticks(yticks)

    if yticklabels is not None:
        ax.set_yticklabels(yticklabels)

    if tick_params:
        ax.tick_params(**tick_params)

    if top_spine is False:
        ax.spines['top'].set_visible(False)
    elif top_spine is True:
        ax.spines['top'].set_visible(True)

    if right_spine is False:
        ax.spines['right'].set_visible(False)
    elif right_spine is True:
        ax.spines['right'].set_visible(True)

    if bottom_spine is False:
        ax.spines['bottom'].set_visible(False)

    if left_spine is False:
        ax.spines['left'].set_visible(False)
    return


def _add_dist_as_grid(fig: plt.Figure, hist_data, other_axes: plt.Axes,
                      xlabel=None, xlabel_kws=None,  **plot_params):
    """Data point distribution plot for numeric feature"""

    ax = fig.add_axes((0.1, 0.1, 0.8, 0.14), sharex=other_axes)
    process_axis(ax, top_spine=False, xlabel=xlabel, xlabel_kws=xlabel_kws, bottom_spine=False,
                 right_spine=False, left_spine=False)
    ax.yaxis.set_visible(False)  # hide the yaxis
    ax.xaxis.set_visible(False)  # hide the x-axis

    color = plot_params.get('pdp_color', '#1A4E5D')
    ax.plot(hist_data, [1] * len(hist_data), '|', color=color, markersize=20)
    return


def _add_surface(ax, x0, x1, pd_vals, cmap, feature0, feature2, **kwargs):

    ax.plot_surface(x0, x1, pd_vals, cmap=cmap, **kwargs)

    ax.set_xlabel(feature0, fontsize=11)
    ax.set_ylabel(feature2, fontsize=11)
    ax.set_zlabel(f"E[f(x) | {feature0} {feature2} ]", fontsize=11)

    return




try:
    import torch
except (ImportError, ModuleNotFoundError):
    torch = None

class HSGLayer(torch.nn.Module):
    def __init__(self, n_units, init_gates_closed):
        super(HSGLayer, self).__init__()
        self.W_R = torch.nn.Linear(n_units, n_units, bias=False)
        self.W_F = torch.nn.Linear(n_units, n_units)
        if init_gates_closed:
            self.W_F.bias = torch.nn.Parameter(torch.Tensor([-2.5]*n_units).cuda())

    def forward(self, s_l_t, s_prime_tm1):
        g = torch.sigmoid(self.W_R(s_prime_tm1) + self.W_F(s_l_t))
        s_prime_t = g*s_prime_tm1 + (1 - g)*s_l_t
        return s_prime_t


class RHNCell(torch.nn.Module):
    def __init__(self, in_feats, n_units, rec_depth=3, couple_gates=True,
                 use_hsg=False, init_gates_closed=False):
        super(RHNCell, self).__init__()
        self.rec_depth = rec_depth
        self.in_feats = in_feats
        self.n_units = n_units
        self.couple_gates = couple_gates
        self.use_HSG = use_hsg
        self.W_H = torch.nn.Linear(in_feats, n_units, bias=False)
        self.W_T = torch.nn.Linear(in_feats, n_units, bias=False)
        if not couple_gates:
            self.W_C = torch.nn.Linear(in_feats, n_units, bias=False)
        self.R_H = torch.nn.ModuleList([torch.nn.Linear(n_units, n_units) for _ in range(rec_depth)])
        self.R_T = torch.nn.ModuleList([torch.nn.Linear(n_units, n_units) for _ in range(rec_depth)])
        if not couple_gates:
            self.R_C = torch.nn.ModuleList([torch.nn.Linear(n_units, n_units) for _ in range(rec_depth)])

        if use_hsg:
            self.HSG = HSGLayer(n_units, init_gates_closed)

        if init_gates_closed:
            for l in range(rec_depth):
                self.R_T[l].bias = torch.nn.Parameter(torch.Tensor([-2.5] * n_units).cuda())
                if not couple_gates:
                    self.R_C[l].bias = torch.nn.Parameter(torch.Tensor([-2.5] * n_units).cuda())

    def forward(self, x, s):
        if self.use_HSG:
            s_prime_tm1 = s
        preds = []
        for l in range(self.rec_depth):
            if l == 0:
                h_l_t = torch.tanh(self.W_H(x) + self.R_H[l](s))
                t_l_t = torch.sigmoid(self.W_T(x) + self.R_T[l](s))
                if not self.couple_gates:
                    c_l_t = torch.sigmoid(self.W_C(x) + self.R_C[l](s))
            else:
                h_l_t = torch.tanh(self.R_H[l](s))
                t_l_t = torch.sigmoid(self.R_T[l](s))
                if not self.couple_gates:
                    c_l_t = torch.sigmoid(self.R_C[l](s))

            if not self.couple_gates:
                s = h_l_t * t_l_t + c_l_t * s
            else:
                s = h_l_t * t_l_t + (1 - t_l_t) * s
            preds.append(s)

        if self.use_HSG:
            s = self.HSG(s, s_prime_tm1)
            preds.pop()
            preds.append(s)
        preds = torch.stack(preds)
        return s, preds


class RHN(torch.nn.Module):
    def __init__(self, in_feats, out_feats, n_units=32, rec_depth=3, couple_gates=True, use_hsg=False,
                 init_gates_closed=False, use_batch_norm=False):
        super(RHN, self).__init__()
        assert rec_depth > 0
        self.rec_depth = rec_depth
        self.in_feats = in_feats
        self.n_units = n_units
        self.init_gates_closed = init_gates_closed
        self.couple_gates = couple_gates
        self.use_HSG = use_hsg
        self.use_batch_norm = use_batch_norm
        self.RHNCell = RHNCell(in_feats, n_units, rec_depth, couple_gates=couple_gates,
                               use_hsg=use_hsg, init_gates_closed=init_gates_closed)
        if use_batch_norm:
            self.bn_x = torch.nn.BatchNorm1d(in_feats)
            self.bn_s = torch.nn.BatchNorm1d(n_units)

    def forward(self, x):
        s = torch.zeros(x.shape[0], self.n_units).cuda()
        preds = []
        highway_states = []
        for t in range(x.shape[1]):
            if self.use_batch_norm:
                x_inp = self.bn_x(x[:, t, :])
                s = self.bn_s(s)
            else:
                x_inp = x[:, t, :]
            s, all_s = self.RHNCell(x_inp, s)
            preds.append(s)
            highway_states.append(all_s)
        preds = torch.stack(preds)
        preds = preds.permute(1, 0, 2)
        highway_states = torch.stack(highway_states)
        highway_states = highway_states.permute(2, 0, 3, 1)
        out = preds

        return out, highway_states


class ConvBlock(torch.nn.Module):
    def __init__(self, timesteps, in_channels, n_filters=32, filter_size=5):
        super(ConvBlock, self).__init__()
        padding1 = self._calc_padding(timesteps, filter_size)
        self.conv = torch.nn.Conv1d(in_channels, n_filters, filter_size, padding=padding1)
        self.relu = torch.nn.ReLU()
        self.maxpool = torch.nn.AdaptiveMaxPool1d(timesteps)
        self.zp = torch.nn.ConstantPad1d((1, 0), 0)

    def _calc_padding(self, lin, kernel, stride=1, dilation=1):
        p = int(((lin - 1) * stride + 1 + dilation * (kernel - 1) - lin) / 2)
        return p

    def forward(self, x):
        x = x.permute(0, 2, 1)
        x = self.conv(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = x.permute(0, 2, 1)
        return x


class HARHN(torch.nn.Module):
    def __init__(self, n_conv_layers, lookback, in_feats, target_feats, n_units_enc=32, n_units_dec=32,
                 enc_input_size=32,
                 rec_depth=3,
                 use_predicted_output=True,
                 out_feats=1, n_filters=32, filter_size=5):
        super(HARHN, self).__init__()
        assert n_conv_layers > 0
        self.n_convs = n_conv_layers
        self.n_units_enc = n_units_enc
        self.n_units_dec = n_units_dec
        self.rec_depth = rec_depth
        self.T = lookback
        self.use_predicted_output = use_predicted_output
        self.convs = torch.nn.ModuleList([ConvBlock(lookback, in_feats, n_filters=n_filters,
                                              filter_size=filter_size) if i == 0 else ConvBlock(lookback, n_filters,
                                                                                                n_filters=n_filters,
                                                                                                filter_size=filter_size)
                                    for i in range(n_conv_layers)])
        self.conv_to_enc = torch.nn.Linear(n_filters, enc_input_size)
        self.RHNEncoder = RHN(enc_input_size, out_feats=n_units_enc, n_units=n_units_enc, rec_depth=rec_depth)
        self.RHNDecoder = RHNCell(target_feats, n_units_dec, rec_depth=rec_depth)
        self.T_k = torch.nn.ModuleList([torch.nn.Linear(n_units_dec, n_units_enc, bias=False) for _ in range(self.rec_depth)])
        self.U_k = torch.nn.ModuleList([torch.nn.Linear(n_units_enc, n_units_enc) for _ in range(self.rec_depth)])
        self.v_k = torch.nn.ModuleList([torch.nn.Linear(n_units_enc, 1) for _ in range(self.rec_depth)])
        self.W_tilda = torch.nn.Linear(target_feats, target_feats, bias=False)
        self.V_tilda = torch.nn.Linear(rec_depth * n_units_enc, target_feats)
        self.W = torch.nn.Linear(n_units_dec, target_feats)
        self.V = torch.nn.Linear(rec_depth * n_units_enc, target_feats)

    def forward(self, x, y_prev_t):
        if self.use_predicted_output:
            y_prev_t = y_prev_t[0:x.shape[0]]
        for conv in range(self.n_convs):
            x = self.convs[conv](x)
        x = self.conv_to_enc(x)
        x, h_t_l = self.RHNEncoder(x)  # h_T_L.shape = (batch_size, T, n_units_enc, rec_depth)
        s = torch.zeros(x.shape[0], self.n_units_dec).cuda()
        for t in range(self.T):
            s_rep = s.unsqueeze(1)
            s_rep = s_rep.repeat(1, self.T, 1)
            d_t = []
            for k in range(self.rec_depth):
                h_t_k = h_t_l[..., k]
                _ = self.U_k[k](h_t_k)
                _ = self.T_k[k](s_rep)
                e_t_k = self.v_k[k](torch.tanh(self.T_k[k](s_rep) + self.U_k[k](h_t_k)))
                alpha_t_k = torch.softmax(e_t_k, 1)
                d_t_k = torch.sum(h_t_k * alpha_t_k, dim=1)
                d_t.append(d_t_k)
            d_t = torch.cat(d_t, dim=1)
            if self.use_predicted_output:
                _, s, y_prev_t = self._last_v2(y_prev_t, d_t, s)
            else:
                _, s, _ = self._last_v1(y_prev_t, d_t, s, t)

        y_t = self.W(s) + self.V(d_t)
        return y_t, y_prev_t

    def _last_v2(self, y_prev_t, d_t, s):
        y_tilda_t = self.W_tilda(y_prev_t) + self.V_tilda(d_t)
        s, _ = self.RHNDecoder(y_tilda_t, s)
        y_prev_t = self.W(s) + self.V(d_t)
        return y_tilda_t, s, y_prev_t

    def _last_v1(self, y_prev_t, d_t, s, t):
        y_tilda_t = self.W_tilda(y_prev_t) + self.V_tilda(d_t)
        s, _ = self.RHNDecoder(y_tilda_t[:, t, :], s)

        return y_tilda_t, s, y_prev_t


import json
import os
import glob
import random
from typing import Union

try:
    import xarray as xr
except ModuleNotFoundError:
    xr = None

import numpy as np
import pandas as pd

from .datasets import Datasets
from .utils import check_attributes, download, sanity_check
from ai4water.utils.utils import dateandtime_now

try:  # shapely may not be installed, as it may be difficult to isntall and is only needed for plotting data.
    from ai4water.preprocessing.spatial_utils import plot_shapefile
except ModuleNotFoundError:
    plot_shapefile = None

# directory separator
SEP = os.sep


def gb_message():
    link = "https://doi.org/10.5285/8344e4f3-d2ea-44f5-8afa-86d2987543a9"
    raise ValueError(f"Dwonlaoad the data from {link} and provide the directory "
                     f"path as dataset=Camels(data=data)")


class Camels(Datasets):

    """
    Get CAMELS dataset.
    This class first downloads the CAMELS dataset if it is not already downloaded.
    Then the selected attribute for a selected id are fetched and provided to the
    user using the method `fetch`.

    Attributes
    -----------
    - ds_dir str/path: diretory of the dataset
    - dynamic_features list: tells which dynamic attributes are available in
      this dataset
    - static_features list: a list of static attributes.
    - static_attribute_categories list: tells which kinds of static attributes
      are present in this category.

    Methods
    ---------
    - stations : returns name/id of stations for which the data (dynamic attributes)
        exists as list of strings.
    - fetch : fetches all attributes (both static and dynamic type) of all
            station/gauge_ids or a speficified station. It can also be used to
            fetch all attributes of a number of stations ids either by providing
            their guage_id or  by just saying that we need data of 20 stations
            which will then be chosen randomly.
    - fetch_dynamic_features :
            fetches speficied dynamic attributes of one specified station. If the
            dynamic attribute is not specified, all dynamic attributes will be
            fetched for the specified station. If station is not specified, the
            specified dynamic attributes will be fetched for all stations.
    - fetch_static_features :
            works same as `fetch_dynamic_features` but for `static` attributes.
            Here if the `category` is not specified then static attributes of
            the specified station for all categories are returned.
        stations : returns list of stations
    """

    DATASETS = {
        'CAMELS-BR': {'url': "https://zenodo.org/record/3964745#.YA6rUxZS-Uk",
                      },
        'CAMELS-GB': {'url': gb_message},
    }

    def stations(self):
        raise NotImplementedError

    def _read_dynamic_from_csv(self, stations, dynamic_features, st=None, en=None):
        raise NotImplementedError

    def fetch_static_features(self, station, features):
        raise NotImplementedError

    @property
    def start(self):  # start of data
        raise NotImplementedError

    @property
    def end(self):  # end of data
        raise NotImplementedError

    @property
    def dynamic_features(self) -> list:
        raise NotImplementedError

    def _check_length(self, st, en):
        if st is None:
            st = self.start
        if en is None:
            en = self.end
        return st, en

    def to_ts(self, static, st, en, as_ts=False, freq='D'):

        st, en = self._check_length(st, en)

        if as_ts:
            idx = pd.date_range(st, en, freq=freq)
            static = pd.DataFrame(np.repeat(static.values, len(idx), axis=0), index=idx,
                                  columns=static.columns)
            return static
        else:
            return static

    @property
    def camels_dir(self):
        """Directory where all camels datasets will be saved. This will under
         datasets directory"""
        return os.path.join(self.base_ds_dir, "CAMELS")

    @property
    def ds_dir(self):
        """Directory where a particular dataset will be saved. """
        return self._ds_dir

    @ds_dir.setter
    def ds_dir(self, x):
        if x is None:
            x = os.path.join(self.camels_dir, self.__class__.__name__)

        if not os.path.exists(x):
            os.makedirs(x)
        # sanity_check(self.name, x)
        self._ds_dir = x

    def fetch(self,
              stations: Union[str, list, int, float, None] = None,
              dynamic_features: Union[list, str, None] = 'all',
              static_features: Union[str, list, None] = None,
              st: Union[None, str] = None,
              en: Union[None, str] = None,
              as_dataframe:bool = False,
              **kwargs
              ) -> Union[dict, pd.DataFrame]:
        """
        Fetches the attributes of one or more stations.

        Arguments:
            stations : if string, it is supposed to be a station name/gauge_id.
                If list, it will be a list of station/gauge_ids. If int, it will
                be supposed that the user want data for this number of
                stations/gauge_ids. If None (default), then attributes of all
                available stations. If float, it will be supposed that the user
                wants data of this fraction of stations.
            dynamic_features : If not None, then it is the attributes to be
                fetched. If None, then all available attributes are fetched
            static_features : list of static attributes to be fetches. None
                means no static attribute will be fetched.
            st : starting date of data to be returned. If None, the data will be
                returned from where it is available.
            en : end date of data to be returned. If None, then the data will be
                returned till the date data is available.
            as_dataframe : whether to return dynamic attributes as pandas
                dataframe or as xarray dataset.
            kwargs : keyword arguments to read the files

        returns:
            If both static  and dynamic features are obtained then it returns a
            dictionary whose keys are station/gauge_ids and values are the
            attributes and dataframes.
            Otherwise either dynamic or static features are returned.

        Examples
        --------
        >>> aus = CAMELS_AUS()
        >>> # get data of 10% of stations
        >>> df = aus.fetch(stations=0.1, static_features=None, as_dataframe=True)

        """
        if isinstance(stations, int):
            # the user has asked to randomly provide data for some specified number of stations
            stations = random.sample(self.stations(), stations)
        elif isinstance(stations, list):
            pass
        elif isinstance(stations, str):
            stations = [stations]
        elif isinstance(stations, float):
            num_stations = int(len(self.stations()) * stations)
            stations = random.sample(self.stations(), num_stations)
        elif stations is None:
            # fetch for all stations
            stations = self.stations()
        else:
            raise TypeError(f"Unknown value provided for stations {stations}")

        if xr is None:
            raise ModuleNotFoundError("modeule xarray must be installed to use `datasets` module")

        return self.fetch_stations_attributes(stations,
                                              dynamic_features,
                                              static_features,
                                              st=st,
                                              en=en,
                                              as_dataframe=as_dataframe,
                                              **kwargs)

    def _maybe_to_netcdf(self, fname: str):
        self.dyn_fname = os.path.join(self.ds_dir, f'{fname}.nc')
        if not os.path.exists(self.dyn_fname):
            # saving all the data in netCDF file using xarray
            print(f'converting data to netcdf format for faster io operations')
            data = self.fetch(static_features=None)
            data_vars = {}
            coords = {}
            for k, v in data.items():
                data_vars[k] = (['time', 'dynamic_features'], v)
                index = v.index
                index.name = 'time'
                coords = {
                    'dynamic_features': list(v.columns),
                    'time': index
                }
            xds = xr.Dataset(
                data_vars=data_vars,
                coords=coords,
                attrs={'date': f"create on {dateandtime_now()}"}
            )
            xds.to_netcdf(self.dyn_fname)

    def fetch_stations_attributes(self,
                                  stations: list,
                                  dynamic_features='all',
                                  static_features=None,
                                  st=None,
                                  en=None,
                                  as_dataframe: bool = False,
                                  **kwargs):
        """Reads attributes of more than one stations.
        Arguments:
            stations : list of stations for which data is to be fetched.
            dynamic_features : list of dynamic attributes to be fetched.
                if 'all', then all dynamic attributes will be fetched.
            static_features : list of static attributes to be fetched.
                If `all`, then all static attributes will be fetched. If None,
                then no static attribute will be fetched.
            st : start of data to be fetched.
            en : end of data to be fetched.
            as_dataframe : whether to return the data as pandas dataframe. default
                is xr.dataset object
            kwargs dict: additional keyword arguments
        Returns:
            Dynamic and static features of multiple stations. Dynamic features
            are by default returned as xr.Dataset unless `as_dataframe` is True, in
            such a case, it is a pandas dataframe with multiindex. If xr.Dataset,
            it consists of `data_vars` equal to number of stations and for each
            station, the `DataArray` is of dimensions (time, dynamic_features).
            where `time` is defined by `st` and `en` i.e length of `DataArray`.
            In case, when the returned object is pandas DataFrame, the first index
            is `time` and second index is `dyanamic_features`. Static attributes
            are always returned as pandas DataFrame and have following shape
            `(stations, static_features). If `dynamic_features` is None,
            then they are not returned and the returned value only consists of
            static features. Same holds true for `static_features`.
            If both are not None, then the returned type is a dictionary with
            `static` and `dynamic` keys.

        Raises:
            ValueError, if both dynamic_features and static_features are None
            """
        st, en = self._check_length(st, en)

        if dynamic_features is not None:

            dynamic_features = check_attributes(dynamic_features, self.dynamic_features)

            if not os.path.exists(self.dyn_fname):
                # read from csv files
                # following code will run only once when fetch is called inside init method
                dyn = self._read_dynamic_from_csv(stations, dynamic_features, st=st, en=en)

            else:
                dyn = xr.load_dataset(self.dyn_fname)  # daataset
                dyn = dyn[stations].sel(dynamic_features=dynamic_features, time=slice(st, en))
                if as_dataframe:
                    dyn = dyn.to_dataframe(['time', 'dynamic_features'])

            if static_features is not None:
                static = self.fetch_static_features(stations, static_features)
                stns = {'dynamic': dyn, 'static': static}
            else:
                stns = dyn

        elif static_features is not None:

            return self.fetch_static_features(stations, static_features)

        else:
            raise ValueError

        return stns

    def fetch_dynamic_features(self,
                               stn_id,
                               attributes='all',
                               st=None,
                               en=None,
                               as_dataframe=False
                               ):
        """Fetches all or selected dynamic attributes of one station."""

        assert isinstance(stn_id, str)
        station = [stn_id]
        return self.fetch_stations_attributes(station,
                                              attributes,
                                              None,
                                              st=st,
                                              en=en,
                                              as_dataframe=as_dataframe)

    def fetch_station_attributes(self,
                                 station: str,
                                 dynamic_features: Union[str, list, None] = 'all',
                                 static_features: Union[str, list, None] = None,
                                 as_ts: bool = False,
                                 st: Union[str, None] = None,
                                 en: Union[str, None] = None,
                                 **kwargs) -> pd.DataFrame:
        """
        Fetches attributes for one station.

        Arguments:
            station : station id/gauge id for which the data is to be fetched.
            dynamic_features:
            static_features:
            as_ts : whether static attributes are to be converted into a time
                series or not. If yes then the returned time series will be of
                same length as that of dynamic attribtues.
            st : starting point from which the data to be fetched. By default
                the data will be fetched from where it is available.
            en : end point of data to be fetched. By default the dat will be fetched

        Return: DataFrame
            dataframe if as_ts is True else it returns a dictionary of static and
                dynamic attributes for a station/gauge_id
            """
        st, en = self._check_length(st, en)

        station_df = pd.DataFrame()
        if dynamic_features:
            dynamic = self.fetch_dynamic_features(station, dynamic_features, st=st,
                                                  en=en, **kwargs)
            station_df = pd.concat([station_df, dynamic])

            if static_features is not None:
                static = self.fetch_static_features(station, static_features)

                if as_ts:
                    station_df = pd.concat([station_df, static], axis=1)
                else:
                    station_df = {'dynamic': station_df, 'static': static}

        elif static_features is not None:
            station_df = self.fetch_static_features(station, static_features)

        return station_df


class LamaH(Camels):
    """
    Large-Sample Data for Hydrology and Environmental Sciences for Central Europe
    from     url = "https://zenodo.org/record/4609826#.YFNp59zt02w"
    paper: https://essd.copernicus.org/preprints/essd-2021-72/
    """
    url = "https://zenodo.org/record/4609826#.YFNp59zt02w"
    _data_types = ['total_upstrm', 'diff_upstrm_all', 'diff_upstrm_lowimp'
                   ]
    time_steps = ['daily', 'hourly'
                  ]

    static_attribute_categories = ['']

    def __init__(self, *,
                 time_step: str,
                 data_type: str,
                 **kwargs
                 ):

        """
        Arguments:
            time_step : possible values are `daily` or `hourly`
            data_type : possible values are `total_upstrm`, `diff_upstrm_all`
                or 'diff_upstrm_lowimp'
        """

        assert time_step in self.time_steps, f"invalid time_step {time_step} given"
        assert data_type in self._data_types, f"invalid data_type {data_type} given."

        self.time_step = time_step
        self.data_type = data_type

        super().__init__(**kwargs)

        self._download()

        fpath = os.path.join(self.ds_dir, 'lamah_diff_upstrm_lowimp_hourly_dyn.nc')

        _data_types = self._data_types if self.time_step == 'daily' else ['total_upstrm']

        if not os.path.exists(fpath):
            for dt in _data_types:
                for ts in self.time_steps:

                    self.time_step = ts
                    self.data_type = dt
                    fname = f"lamah_{dt}_{ts}_dyn"
                    self._maybe_to_netcdf(fname)

            self.time_step = time_step
            self.data_type = data_type

        self.dyn_fname = os.path.join(self.ds_dir,
                                      f'lamah_{data_type}_{time_step}_dyn.nc')

    @property
    def dynamic_features(self):
        station = self.stations()[0]
        df = self.read_ts_of_station(station)
        return df.columns.to_list()

    @property
    def static_features(self) -> list:
        fname = os.path.join(self.data_type_dir,
                             f'1_attributes{SEP}Catchment_attributes.csv')
        df = pd.read_csv(fname, sep=';', index_col='ID')
        return df.columns.to_list()

    @property
    def ds_dir(self):
        """Directory where a particular dataset will be saved. """
        return os.path.join(self.camels_dir, self.name)

    @property
    def data_type_dir(self):
        directory = 'CAMELS_AT'
        if self.time_step == 'hourly':
            directory = 'CAMELS_AT1'  # todo, use it only for hourly, daily is causing errors
        # self.ds_dir/CAMELS_AT/data_type_dir
        f = [f for f in os.listdir(os.path.join(self.ds_dir, directory)) if self.data_type in f][0]
        return os.path.join(self.ds_dir, f'{directory}{SEP}{f}')

    def stations(self) -> list:
        # assuming file_names of the format ID_{stn_id}.csv
        _dirs = os.listdir(os.path.join(self.data_type_dir,
                                        f'2_timeseries{SEP}{self.time_step}'))
        s = [f.split('_')[1].split('.csv')[0] for f in _dirs]
        return s

    def _read_dynamic_from_csv(self,
                               stations,
                               dynamic_features: Union[str, list] = 'all',
                               st=None,
                               en=None,
                               ):
        """Reads attributes of one station"""

        stations_attributes = {}

        for station in stations:

            station_df = pd.DataFrame()

            if dynamic_features is not None:

                dynamic_df = self.read_ts_of_station(station)

                station_df = pd.concat([station_df, dynamic_df])

            stations_attributes[station] = station_df

        return stations_attributes

    def fetch_static_features(self,
                              station: Union[str, list],
                              features=None
                              ) -> pd.DataFrame:

        fname = os.path.join(self.data_type_dir,
                             f'1_attributes{SEP}Catchment_attributes.csv')

        df = pd.read_csv(fname, sep=';', index_col='ID')

        # if features is not None:
        static_features = check_attributes(features, self.static_features)

        df = df[static_features]

        if isinstance(station, list):
            stations = [str(i) for i in station]
        elif isinstance(station, int):
            stations = str(station)
        else:
            stations = station

        df.index = df.index.astype(str)
        df = df.loc[stations]
        if isinstance(df, pd.Series):
            df = pd.DataFrame(df).transpose()

        return df

    def read_ts_of_station(self, station) -> pd.DataFrame:
        # read a file containing timeseries data for one station
        fname = os.path.join(self.data_type_dir,
                             f'2_timeseries{SEP}{self.time_step}{SEP}ID_{station}.csv')

        df = pd.read_csv(fname, sep=';')

        if self.time_step == 'daily':
            periods = pd.PeriodIndex(year=df["YYYY"], month=df["MM"], day=df["DD"],
                                     freq="D")
            df.index = periods.to_timestamp()
        else:
            periods = pd.PeriodIndex(year=df["YYYY"],
                                     month=df["MM"], day=df["DD"], hour=df["hh"],
                                     minute=df["mm"], freq="H")
            df.index = periods.to_timestamp()

        # remove the cols specifying index
        [df.pop(item) for item in ['YYYY', 'MM', 'DD', 'hh', 'mm'] if item in df]
        return df

    @property
    def start(self):
        return "19810101"

    @property
    def end(self):
        return "20191231"


class HYSETS(Camels):
    """
    database for hydrometeorological modeling of 14,425 North American watersheds
    from 1950-2018 following the work of Arsenault_ et al., 2020
    The user must manually download the files, unpack them and provide
    the `path` where these files are saved.

    This data comes with multiple sources. Each source having one or more dynamic_features
    Following data_source are available.

    +---------------+------------------------------+
    |sources        | dynamic_features             |
    |===============|==============================|
    |SNODAS_SWE     | dscharge, swe                |
    +---------------+------------------------------+
    |SCDNA          | discharge, pr, tasmin, tasmax|
    +---------------+------------------------------+
    |nonQC_stations | discharge, pr, tasmin, tasmax|
    +---------------+------------------------------+
    |Livneh         | discharge, pr, tasmin, tasmax|
    +---------------+------------------------------+
    |ERA5           | discharge, pr, tasmax, tasmin|
    +---------------+------------------------------+
    |ERAS5Land_SWE  | discharge, swe               |
    +---------------+------------------------------+
    |ERA5Land       | discharge, pr, tasmax, tasmin|
    +---------------+------------------------------+

    all sources contain one or more following dynamic_features
    with following shapes

    +----------------------------+------------------+
    |dynamic_features            |      shape       |
    |============================|==================|
    |time                        |   (25202,)       |
    +----------------------------+------------------+
    |watershedID                 |   (14425,)       |
    +----------------------------+------------------+
    |drainage_area               |   (14425,)       |
    +----------------------------+------------------+
    |drainage_area_GSIM          |   (14425,)       |
    +----------------------------+------------------+
    |flag_GSIM_boundaries        |   (14425,)       |
    +----------------------------+------------------+
    |flag_artificial_boundaries  |   (14425,)       |
    +----------------------------+------------------+
    |centroid_lat                |   (14425,)       |
    +----------------------------+------------------+
    |centroid_lon                |   (14425,)       |
    +----------------------------+------------------+
    |elevation                   |   (14425,)       |
    +----------------------------+------------------+
    |slope                       |   (14425,)       |
    +----------------------------+------------------+
    |discharge                   |   (14425, 25202) |
    +----------------------------+------------------+
    |pr                          |   (14425, 25202) |
    +----------------------------+------------------+
    |tasmax                      |   (14425, 25202) |
    +----------------------------+------------------+
    |tasmin                      |   (14425, 25202) |
    +----------------------------+------------------+

    Examples
    --------
    >>> dataset = HYSETS(path="path/to/HYSETS")
    >>> df = dataset.fetch(0.01, as_dataframe=True) # 1% of stations

    .. _Arsenault:
        https://doi.org/10.1038/s41597-020-00583-2
    """
    doi = "https://doi.org/10.1038/s41597-020-00583-2"
    url = "https://osf.io/rpc3w/"
    Q_SRC = ['ERA5', 'ERA5Land', 'ERA5Land_SWE', 'Livneh', 'nonQC_stations', 'SCDNA', 'SNODAS_SWE']
    SWE_SRC = ['ERA5Land_SWE', 'SNODAS_SWE']
    OTHER_SRC = [src for src in Q_SRC if src not in  ['ERA5Land_SWE', 'SNODAS_SWE']]
    dynamic_features = ['discharge', 'swe', 'tasmin', 'tasmax', 'pr']

    def __init__(self,
                 path: str,
                 swe_source: str = "SNODAS_SWE",
                 discharge_source: str = "ERA5",
                 tasmin_source: str = "ERA5",
                 tasmax_source: str = "ERA5",
                 pr_source: str = "ERA5",
                 **kwargs
                 ):
        """
        Arguments:
            path : path where all the data files are saved.
            swe_source : source of swe data.
            discharge_source : source of discharge data
            tasmin_source : source of tasmin data
            tasmax_source : source of tasmax data
            pr_source : source of pr data
            kwargs : arguments for `Camels` base class

        """

        assert swe_source in self.SWE_SRC, f'source must be one of {self.SWE_SRC}'
        assert discharge_source in self.Q_SRC, f'source must be one of {self.Q_SRC}'
        assert tasmin_source in self.OTHER_SRC, f'source must be one of {self.OTHER_SRC}'
        assert tasmax_source in self.OTHER_SRC, f'source must be one of {self.OTHER_SRC}'
        assert pr_source in self.OTHER_SRC, f'source must be one of {self.OTHER_SRC}'

        self.sources = {
            'swe': swe_source,
            'discharge': discharge_source,
            'tasmin': tasmin_source,
            'tasmax': tasmax_source,
            'pr': pr_source
        }

        super().__init__(**kwargs)

        self.ds_dir = path

        fpath = os.path.join(self.ds_dir, 'hysets_dyn.nc')
        if not os.path.exists(fpath):
            self._maybe_to_netcdf('hysets_dyn')

    def _maybe_to_netcdf(self, fname:str):
        # todo saving as one file takes very long time
        oneD_vars = []
        twoD_vars = []

        for src in self.Q_SRC:
            xds = xr.open_dataset(os.path.join(self.ds_dir, f'HYSETS_2020_{src}.nc'))

            for var in xds.variables:
                print(f'getting {var} from source {src} ')

                if len(xds[var].data.shape) > 1:
                    xar = xds[var]
                    xar.name = f"{xar.name}_{src}"
                    twoD_vars.append(xar)
                else:
                    xar = xds[var]
                    xar.name = f"{xar.name}_{src}"
                    oneD_vars.append(xar)

        oneD_xds = xr.merge(oneD_vars)
        twoD_xds = xr.merge(twoD_vars)
        oneD_xds.to_netcdf(os.path.join(self.ds_dir, "hysets_static.nc"))
        twoD_xds.to_netcdf(os.path.join(self.ds_dir, "hysets_dyn.nc"))

        return

    @property
    def ds_dir(self):
        return self._ds_dir

    @ds_dir.setter
    def ds_dir(self, x):
        sanity_check('HYSETS', x)
        self._ds_dir = x

    @property
    def static_features(self):
        df = self.read_static_data()
        return df.columns.to_list()

    def stations(self) -> list:
        return self.read_static_data().index.to_list()

    @property
    def start(self):
        return "19500101"

    @property
    def end(self):
        return "20181231"

    def fetch_stations_attributes(
            self,
            stations: list,
            dynamic_features: Union[str, list, None] = 'all',
            static_features: Union[str, list, None] = None,
            st=None,
            en=None,
            as_dataframe: bool = False,
            **kwargs
    ):

        stations = check_attributes(stations, self.stations())
        stations = [int(stn) for stn in stations]

        if dynamic_features is not None:

            dyn = self._fetch_dynamic_features(stations=stations,
                                               dynamic_features=dynamic_features,
                                               as_dataframe=as_dataframe,
                                               **kwargs
                                               )

            if static_features is not None:  # we want both static and dynamic
                to_return = {}
                static = self._fetch_static_features(station=stations,
                                                     static_features=static_features,
                                                     **kwargs
                                                     )
                to_return['static'] = static
                to_return['dynamic'] = dyn
            else:
                to_return = dyn

        elif static_features is not None:
            # we want only static
            to_return = self._fetch_static_features(
                station=stations,
                static_features=static_features,
                **kwargs
            )
        else:
            raise ValueError

        return to_return

    def fetch_dynamic_features(
            self,
            station,
            dynamic_features='all',
            st=None,
            en=None,
            as_dataframe=False
    ):
        """Fetches dynamic attributes of one station."""
        station = [int(station)]
        return self._fetch_dynamic_features(
            stations=station,
            dynamic_features=dynamic_features,
            st=st,
            en=en,
            as_dataframe=as_dataframe
        )

    def _fetch_dynamic_features(
            self,
            stations: list,
            dynamic_features='all',
            st=None,
            en=None,
            as_dataframe=False,
            as_ts=False
    ):
        """Fetches dynamic attributes of station."""
        st, en = self._check_length(st, en)
        attrs = check_attributes(dynamic_features, self.dynamic_features)

        stations = np.subtract(stations, 1).tolist()
        # maybe we don't need to read all variables
        sources = {k: v for k, v in self.sources.items() if k in attrs}

        # original .nc file contains datasets with dynamic and static features as data_vars
        # however, for uniformity of this API and easy usage, we want a Dataset to have
        # station names/gauge_ids as data_vars and each data_var has
        # dimension (time, dynamic_variables)
        # Therefore, first read all data for each station from .nc file
        # then rearrange it.
        # todo, this operation is slower because of `to_dataframe`
        # also doing this removes all the metadata
        x = {}
        f = os.path.join(self.ds_dir, "hysets_dyn.nc")
        xds = xr.open_dataset(f)
        for stn in stations:
            xds1 = xds[[f'{k}_{v}' for k, v in sources.items()]].sel(watershed=stn, time=slice(st, en))
            xds1 = xds1.rename_vars({f'{k}_{v}': k for k, v in sources.items()})
            x[stn] = xds1.to_dataframe(['time'])
        xds = xr.Dataset(x)
        xds = xds.rename_dims({'dim_1': 'dynamic_features'})
        xds = xds.rename_vars({'dim_1': 'dynamic_features'})

        if as_dataframe:
            return xds.to_dataframe(['time', 'dynamic_features'])

        return xds

    def _fetch_static_features(
            self,
            station,
            static_features: Union[str, list] = 'all',
            st=None,
            en=None,
            as_ts=False
    ):

        df = self.read_static_data()

        static_features = check_attributes(static_features, self.static_features)

        if isinstance(station, str):
            station = [station]
        elif isinstance(station, int):
            station = [str(station)]
        elif isinstance(station, list):
            station = [str(stn) for stn in station]
        else:
            raise ValueError

        return self.to_ts(df.loc[station][static_features], st=st, en=en, as_ts=as_ts)

    def fetch_static_features(self,
                              station,
                              features='all',
                              st=None,
                              en=None,
                              as_ts=False
                              ) -> pd.DataFrame:

        return self._fetch_static_features(station, features, st, en, as_ts)

    def read_static_data(self):
        fname = os.path.join(self.ds_dir, 'HYSETS_watershed_properties.txt')
        static_df = pd.read_csv(fname, index_col='Watershed_ID', sep=';')
        static_df.index = static_df.index.astype(str)
        return static_df


class CAMELS_US(Camels):
    """
    Downloads and processes CAMELS dataset of 671 catchments named as CAMELS
    from https://ral.ucar.edu/solutions/products/camels
    https://doi.org/10.5194/hess-19-209-2015
    """
    DATASETS = ['CAMELS_US']
    url = "https://ral.ucar.edu/sites/default/files/public/product-tool/camels-catchment-attributes-and-meteorology-for-large-sample-studies-dataset-downloads/basin_timeseries_v1p2_metForcing_obsFlow.zip"
    catchment_attr_url = "https://ral.ucar.edu/sites/default/files/public/product-tool/camels-catchment-attributes-and-meteorology-for-large-sample-studies-dataset-downloads/camels_attributes_v2.0.zip"

    folders = {'basin_mean_daymet': f'basin_mean_forcing{SEP}daymet',
               'basin_mean_maurer': f'basin_mean_forcing{SEP}maurer',
               'basin_mean_nldas': f'basin_mean_forcing{SEP}nldas',
               'basin_mean_v1p15_daymet': f'basin_mean_forcing{SEP}v1p15{SEP}daymet',
               'basin_mean_v1p15_nldas': f'basin_mean_forcing{SEP}v1p15{SEP}nldas',
               'elev_bands': f'elev{SEP}daymet',
               'hru': f'hru_forcing{SEP}daymet'}

    dynamic_features = ['dayl(s)', 'prcp(mm/day)', 'srad(W/m2)',
                        'swe(mm)', 'tmax(C)', 'tmin(C)', 'vp(Pa)', 'Flow']

    def __init__(self, data_source='basin_mean_daymet'):

        assert data_source in self.folders, f'allwed data sources are {self.folders.keys()}'
        self.data_source = data_source

        super().__init__("CAMELS_US")

        if os.path.exists(self.ds_dir):
            print(f"dataset is already downloaded at {self.ds_dir}")
        else:
            download(self.url, os.path.join(self.camels_dir, f'CAMELS_US{SEP}CAMELS_US.zip'))
            download(self.catchment_attr_url, os.path.join(self.camels_dir, f"CAMELS_US{SEP}catchment_attrs.zip"))
            self._unzip()

        self.attr_dir = os.path.join(self.ds_dir, f'catchment_attrs{SEP}camels_attributes_v2.0')
        self.dataset_dir = os.path.join(self.ds_dir, f'CAMELS_US{SEP}basin_dataset_public_v1p2')

        self._maybe_to_netcdf('camels_us_dyn')

    @property
    def ds_dir(self):
        """Directory where a particular dataset will be saved. """
        return os.path.join(self.camels_dir, self.name)

    @property
    def start(self):
        return "19800101"

    @property
    def end(self):
        return "20141231"

    @property
    def static_features(self):
        static_fpath = os.path.join(self.ds_dir, 'static_features.csv')
        if not os.path.exists(static_fpath):
            files = glob.glob(f"{os.path.join(self.ds_dir, 'catchment_attrs', 'camels_attributes_v2.0')}/*.txt")
            cols = []
            for f in files:
                _df = pd.read_csv(f, sep=';', index_col='gauge_id', nrows=1)
                cols += list(_df.columns)
        else:
            df = pd.read_csv(static_fpath, index_col='gauge_id', nrows=1)
            cols = list(df.columns)

        return cols

    def stations(self) -> list:
        stns = []
        for _dir in os.listdir(os.path.join(self.dataset_dir, 'usgs_streamflow')):
            cat = os.path.join(self.dataset_dir, f'usgs_streamflow{SEP}{_dir}')
            stns += [fname.split('_')[0] for fname in os.listdir(cat)]

        # remove stations for which static values are not available
        for stn in ['06775500', '06846500', '09535100']:
            stns.remove(stn)

        return stns

    def _read_dynamic_from_csv(self,
                               stations,
                               dynamic_features: Union[str, list] = 'all',
                               st=None,
                               en=None,
                               ):
        dyn = {}
        for station in stations:

            # attributes = check_attributes(dynamic_features, self.dynamic_features)

            assert isinstance(station, str)
            df = None
            df1 = None
            dir_name = self.folders[self.data_source]
            for cat in os.listdir(os.path.join(self.dataset_dir, dir_name)):
                cat_dirs = os.listdir(os.path.join(self.dataset_dir, f'{dir_name}{SEP}{cat}'))
                stn_file = f'{station}_lump_cida_forcing_leap.txt'
                if stn_file in cat_dirs:
                    df = pd.read_csv(os.path.join(self.dataset_dir,
                                                  f'{dir_name}{SEP}{cat}{SEP}{stn_file}'),
                                     sep="\s+|;|:",
                                     skiprows=4,
                                     engine='python',
                                     names=['Year', 'Mnth', 'Day', 'Hr', 'dayl(s)', 'prcp(mm/day)', 'srad(W/m2)',
                                            'swe(mm)', 'tmax(C)', 'tmin(C)', 'vp(Pa)'],
                                     )
                    df.index = pd.to_datetime(df['Year'].map(str) + '-' + df['Mnth'].map(str) + '-' + df['Day'].map(str))

            flow_dir = os.path.join(self.dataset_dir, 'usgs_streamflow')
            for cat in os.listdir(flow_dir):
                cat_dirs = os.listdir(os.path.join(flow_dir, cat))
                stn_file = f'{station}_streamflow_qc.txt'
                if stn_file in cat_dirs:
                    fpath = os.path.join(flow_dir, f'{cat}{SEP}{stn_file}')
                    df1 = pd.read_csv(fpath,  sep="\s+|;|:'",
                                      names=['station', 'Year', 'Month', 'Day', 'Flow', 'Flag'],
                                      engine='python')
                    df1.index = pd.to_datetime(
                        df1['Year'].map(str) + '-' + df1['Month'].map(str) + '-' + df1['Day'].map(str))

            out_df = pd.concat([df[['dayl(s)',
                                    'prcp(mm/day)', 'srad(W/m2)', 'swe(mm)', 'tmax(C)', 'tmin(C)', 'vp(Pa)']],
                                df1['Flow']],
                               axis=1)
            dyn[station] = out_df

        return dyn

    def fetch_static_features(self, station, features):

        attributes = check_attributes(features, self.static_features)

        static_fpath = os.path.join(self.ds_dir, 'static_features.csv')
        if not os.path.exists(static_fpath):
            files = glob.glob(f"{os.path.join(self.ds_dir, 'catchment_attrs', 'camels_attributes_v2.0')}/*.txt")
            static_df = pd.DataFrame()
            for f in files:
                # index should be read as string
                idx = pd.read_csv(f, sep=';', usecols=['gauge_id'], dtype=str)
                _df = pd.read_csv(f, sep=';', index_col='gauge_id')
                _df.index = idx['gauge_id']
                static_df = pd.concat([static_df, _df], axis=1)
            static_df.to_csv(static_fpath, index_label='gauge_id')
        else:   # index should be read as string bcs it has 0s at the start
            idx = pd.read_csv(static_fpath, usecols=['gauge_id'], dtype=str)
            static_df = pd.read_csv(static_fpath, index_col='gauge_id')
            static_df.index = idx['gauge_id']

        static_df.index = static_df.index.astype(str)
        df = static_df.loc[station][attributes]
        if isinstance(df, pd.Series):
            df = pd.DataFrame(df).transpose()

        return df


class CAMELS_BR(Camels):
    """
    Downloads and processes CAMELS dataset of Brazil
    """
    url = "https://zenodo.org/record/3964745#.YA6rUxZS-Uk"

    folders = {'streamflow_m3s': '02_CAMELS_BR_streamflow_m3s',
               'streamflow_mm': '03_CAMELS_BR_streamflow_mm_selected_catchments',
               'simulated_streamflow_m3s': '04_CAMELS_BR_streamflow_simulated',
               'precipitation_cpc': '07_CAMELS_BR_precipitation_cpc',
               'precipitation_mswep': '06_CAMELS_BR_precipitation_mswep',
               'precipitation_chirps': '05_CAMELS_BR_precipitation_chirps',
               'evapotransp_gleam': '08_CAMELS_BR_evapotransp_gleam',
               'evapotransp_mgb': '09_CAMELS_BR_evapotransp_mgb',
               'potential_evapotransp_gleam': '10_CAMELS_BR_potential_evapotransp_gleam',
               'temperature_min': '11_CAMELS_BR_temperature_min_cpc',
               'temperature_mean': '12_CAMELS_BR_temperature_mean_cpc',
               'temperature_max': '13_CAMELS_BR_temperature_max_cpc'
               }

    def __init__(self):

        super().__init__("CAMELS-BR")

        self._download()

        self._maybe_to_netcdf('camels_dyn_br')

    @property
    def ds_dir(self):
        """Directory where a particular dataset will be saved. """
        return os.path.join(self.camels_dir, self.name)

    @property
    def _all_dirs(self):
        """All the folders in the dataset_directory"""
        return [f for f in os.listdir(self.ds_dir) if os.path.isdir(os.path.join(self.ds_dir, f))]

    @property
    def static_dir(self):
        path = None
        for _dir in self._all_dirs:
            if "attributes" in _dir:
                # supposing that 'attributes' axist in only one file/folder in self.ds_dir
                path = os.path.join(self.ds_dir, f'{_dir}{SEP}{_dir}')
        return path

    @property
    def static_files(self):
        all_files = None
        if self.static_dir is not None:
            all_files = glob.glob(f"{self.static_dir}/*.txt")
        return all_files

    @property
    def dynamic_features(self) -> list:
        return list(CAMELS_BR.folders.keys())

    @property
    def static_attribute_categories(self):
        static_attrs = []
        for f in self.static_files:
            ff = str(os.path.basename(f).split('.txt')[0])
            static_attrs.append('_'.join(ff.split('_')[2:]))
        return static_attrs

    @property
    def static_features(self):
        static_fpath = os.path.join(self.ds_dir, 'static_features.csv')
        if not os.path.exists(static_fpath):
            files = glob.glob(f"{os.path.join(self.ds_dir, '01_CAMELS_BR_attributes','01_CAMELS_BR_attributes')}/*.txt")
            cols = []
            for f in files:
                _df = pd.read_csv(f, sep=' ', index_col='gauge_id', nrows=1)
                cols += list(_df.columns)
        else:
            df = pd.read_csv(static_fpath, index_col='gauge_id', nrows=1)
            cols = list(df.columns)

        return cols

    @property
    def start(self):
        return "19800101"

    @property
    def end(self):
        return "20181231"

    def all_stations(self, attribute) -> list:
        """Tells all station ids for which a data of a specific attribute is available."""
        all_files = []
        for _attr, _dir in self.folders.items():
            if attribute in _attr:
                all_files = os.listdir(os.path.join(self.ds_dir, f'{_dir}{SEP}{_dir}'))

        stations = []
        for f in all_files:
            stations.append(str(f.split('_')[0]))

        return stations

    def stations(self, to_exclude=None) -> list:
        """Returns a list of station ids which are common among all dynamic 
        attributes.
        
        Example
        -------
        >>> dataset = CAMELS_BR()
        >>> stations = dataset.stations()
        """
        if to_exclude is not None:
            if not isinstance(to_exclude, list):
                assert isinstance(to_exclude, str)
                to_exclude = [to_exclude]
        else:
            to_exclude = []

        stations = {}
        for dyn_attr in self.dynamic_features:
            if dyn_attr not in to_exclude:
                stations[dyn_attr] = self.all_stations(dyn_attr)

        stns = list(set.intersection(*map(set, list(stations.values()))))
        return stns

    def _read_dynamic_from_csv(self,
                               stations,
                               attributes: Union[str, list] = 'all',
                               st=None,
                               en=None,
                               ):
        """
        returns the dynamic/time series attribute/attributes for one station id.
        
        Example
        -------
        >>> dataset = CAMELS_BR()
        >>> pcp = dataset.fetch_dynamic_features('10500000', 'precipitation_cpc')
        ... # fetch all time series data associated with a station.
        >>> x = dataset.fetch_dynamic_features('51560000', dataset.dynamic_features)
        
        """

        attributes = check_attributes(attributes, self.dynamic_features)

        dyn = {}
        for stn_id in stations:
            # making one separate dataframe for one station
            data = pd.DataFrame()
            for attr, _dir in self.folders.items():

                if attr in attributes:
                    path = os.path.join(self.ds_dir, f'{_dir}{SEP}{_dir}')
                    # supposing that the filename starts with stn_id and has .txt extension.
                    fname = [f for f in os.listdir(path) if f.startswith(str(stn_id)) and f.endswith('.txt')]
                    fname = fname[0]
                    if os.path.exists(os.path.join(path, fname)):
                        df = pd.read_csv(os.path.join(path, fname), sep=' ')
                        df.index = pd.to_datetime(df[['year', 'month', 'day']])
                        df.index.freq = pd.infer_freq(df.index)
                        df = df[st:en]
                        # only read one column which matches the attr
                        # todo, qual_flag maybe important
                        [df.pop(item) for item in df.columns if item != attr]
                        data = pd.concat([data, df], axis=1)
                    else:
                        raise FileNotFoundError(f"file {fname} not found at {path}")

            dyn[stn_id] = data

        return dyn

    def fetch_static_features(self,
                              station,
                              features=None
                              ) -> pd.DataFrame:
        """
        Arguments:
        stn_id int/list:
            station id whose attribute to fetch
        attributes str/list:
            name of attribute to fetch. Default is None, which will return all the
            attributes for a particular station of the specified category.
        index_col_name str:
            name of column containing station names

        as_ts bool:

        Example
        -------
        >>> dataset = Camels('CAMELS-BR')
        >>> df = dataset.fetch_static_features(11500000, 'climate')

        """
        if isinstance(station, int):
            station = [str(station)]
        elif isinstance(station, list):
            station = [str(stn) for stn in station]
        elif isinstance(station, str):
            station = [station]
        else:
            raise ValueError

        attributes = check_attributes(features, self.static_features)

        static_fpath = os.path.join(self.ds_dir, 'static_features.csv')
        if not os.path.exists(static_fpath):
            files = glob.glob(f"{os.path.join(self.ds_dir, '01_CAMELS_BR_attributes','01_CAMELS_BR_attributes')}/*.txt")
            static_df = pd.DataFrame()
            for f in files:
                _df = pd.read_csv(f, sep=' ', index_col='gauge_id')
                static_df = pd.concat([static_df, _df], axis=1)
            static_df.to_csv(static_fpath, index_label='gauge_id')
        else:
            static_df = pd.read_csv(static_fpath, index_col='gauge_id')

        static_df.index = static_df.index.astype(str)

        return pd.DataFrame(static_df.loc[station][attributes])


class CAMELS_GB(Camels):
    """
    This dataset must be manually downloaded by the user.
    The path of the downloaded folder must be provided while initiating this class.
    """
    dynamic_features = ["precipitation", "pet", "temperature", "discharge_spec",
                        "discharge_vol", "peti",
                        "humidity", "shortwave_rad", "longwave_rad", "windspeed"]

    def __init__(self, path=None):
        super().__init__(name="CAMELS-GB")
        self.ds_dir = path

        self._maybe_to_netcdf('camels_gb_dyn')

    @property
    def ds_dir(self):
        """Directory where a particular dataset will be saved. """
        return self._ds_dir

    @ds_dir.setter
    def ds_dir(self, x):
        sanity_check('CAMELS-GB', x)
        self._ds_dir = x

    @property
    def static_attribute_categories(self) -> list:
        attributes = []
        path = os.path.join(self.ds_dir, 'data')
        for f in os.listdir(path):
            if os.path.isfile(os.path.join(path, f)) and f.endswith('csv'):
                attributes.append(f.split('_')[2])

        return attributes

    @property
    def start(self):
        return "19701001"

    @property
    def end(self):
        return "20150930"

    @property
    def static_features(self):
        files = glob.glob(f"{os.path.join(self.ds_dir, 'data')}/*.csv")
        cols = []
        for f in files:
            if 'static_features.csv' not in f:
                df = pd.read_csv(f, nrows=1, index_col='gauge_id')
                cols += (list(df.columns))
        return cols

    def stations(self, to_exclude=None):
        # CAMELS_GB_hydromet_timeseries_StationID_number
        path = os.path.join(self.ds_dir, f'data{SEP}timeseries')
        gauge_ids = []
        for f in os.listdir(path):
            gauge_ids.append(f.split('_')[4])

        return gauge_ids

    def _read_dynamic_from_csv(self,
                               stations,
                               attributes: Union[str, list] = 'all',
                               st=None,
                               en=None,
                               ):
        """Fetches dynamic attribute/attributes of one station."""
        dyn = {}
        for stn_id in stations:
            # making one separate dataframe for one station
            path = os.path.join(self.ds_dir, f"data{SEP}timeseries")
            fname = None
            for f in os.listdir(path):
                if stn_id in f:
                    fname = f
                    break

            df = pd.read_csv(os.path.join(path, fname), index_col= 'date')
            df.index = pd.to_datetime(df.index)
            df.index.freq = pd.infer_freq(df.index)

            dyn[stn_id] = df

        return dyn

    def fetch_static_features(self,
                              station: str,
                              features='all'
                              ) -> pd.DataFrame:
        """Fetches static attributes of one station for one or more category as dataframe."""

        attributes = check_attributes(features, self.static_features)
        static_fname = 'static_features.csv'
        static_fpath = os.path.join(self.ds_dir, 'data', static_fname)
        if os.path.exists(static_fpath):
            static_df = pd.read_csv(static_fpath, index_col='gauge_id')
        else:
            files = glob.glob(f"{os.path.join(self.ds_dir, 'data')}/*.csv")
            static_df = pd.DataFrame()
            for f in files:
                _df = pd.read_csv(f, index_col='gauge_id')
                static_df = pd.concat([static_df, _df], axis=1)
            static_df.to_csv(static_fpath)

        if isinstance(station, str):
            station = [station]
        elif isinstance(station, int):
            station = [str(station)]
        elif isinstance(station, list):
            station = [str(stn) for stn in station]
        else:
            raise ValueError

        static_df.index = static_df.index.astype(str)

        return static_df.loc[station][attributes]


class CAMELS_AUS(Camels):
    """
    Inherits from Camels class. Reads CAMELS-AUS dataset of
    [Fowler et al., 2020](https://doi.org/10.5194/essd-13-3847-2021)
    dataset.

    Examples
    --------
    >>> dataset = CAMELS_AUS()
    >>> df = dataset.fetch(stations=1, as_dataframe=True)
    >>> df.unstack() # the returned dataframe is a multi-indexed dataframe so we have to unstack it
    """
    url = 'https://doi.pangaea.de/10.1594/PANGAEA.921850'
    urls = {
        "01_id_name_metadata.zip": "https://download.pangaea.de/dataset/921850/files/",
        "02_location_boundary_area.zip": "https://download.pangaea.de/dataset/921850/files/",
        "03_streamflow.zip": "https://download.pangaea.de/dataset/921850/files/",
        "04_attributes.zip": "https://download.pangaea.de/dataset/921850/files/",
        "05_hydrometeorology.zip": "https://download.pangaea.de/dataset/921850/files/",
        "CAMELS_AUS_Attributes-Indices_MasterTable.csv": "https://download.pangaea.de/dataset/921850/files/",
        "Units_01_TimeseriesData.pdf": "https://download.pangaea.de/dataset/921850/files/",
        "Units_02_AttributeMasterTable.pdf": "https://download.pangaea.de/dataset/921850/files/",
    }

    folders = {
        'streamflow_MLd': f'03_streamflow{SEP}03_streamflow{SEP}streamflow_MLd',
        'streamflow_MLd_inclInfilled': f'03_streamflow{SEP}03_streamflow{SEP}streamflow_MLd_inclInfilled',
        'streamflow_mmd': f'03_streamflow{SEP}03_streamflow{SEP}streamflow_mmd',

        'et_morton_actual_SILO':  f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}et_morton_actual_SILO',
        'et_morton_point_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}et_morton_point_SILO',
        'et_morton_wet_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}et_morton_wet_SILO',
        'et_short_crop_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}et_short_crop_SILO',
        'et_tall_crop_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}et_tall_crop_SILO',
        'evap_morton_lake_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}evap_morton_lake_SILO',
        'evap_pan_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}evap_pan_SILO',
        'evap_syn_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}02_EvaporativeDemand_timeseries{SEP}evap_syn_SILO',

        'precipitation_AWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}01_precipitation_timeseries{SEP}precipitation_AWAP',
        'precipitation_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}01_precipitation_timeseries{SEP}precipitation_SILO',
        'precipitation_var_SWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}01_precipitation_timeseries{SEP}precipitation_var_AWAP',

        'solarrad_AWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}AWAP{SEP}solarrad_AWAP',
        'tmax_AWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}AWAP{SEP}tmax_AWAP',
        'tmin_AWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}AWAP{SEP}tmin_AWAP',
        'vprp_AWAP': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}AWAP{SEP}vprp_AWAP',

        'mslp_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}mslp_SILO',
        'radiation_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}radiation_SILO',
        'rh_tmax_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}rh_tmax_SILO',
        'rh_tmin_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}rh_tmin_SILO',
        'tmax_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}tmax_SILO',
        'tmin_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}tmin_SILO',
        'vp_deficit_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}vp_deficit_SILO',
        'vp_SILO': f'05_hydrometeorology{SEP}05_hydrometeorology{SEP}03_Other{SEP}SILO{SEP}vp_SILO',
    }

    def __init__(self, path: str = None):
        """
        Arguments:
            path: path where the CAMELS-AUS dataset has been downloaded. This path
                must contain five zip files and one xlsx file. If None, then the
                data will downloaded.
        """
        if path is not None:
            assert isinstance(path, str), f'path must be string like but it is "{path}" of type {path.__class__.__name__}'
            if not os.path.exists(path) or len(os.listdir(path)) < 2:
                raise FileNotFoundError(f"The path {path} does not exist")
        self.ds_dir = path

        super().__init__()
        if not os.path.exists(self.ds_dir):
            os.makedirs(self.ds_dir)

        for _file, url in self.urls.items():
            fpath = os.path.join(self.ds_dir, _file)
            if not os.path.exists(fpath):
                download(url + _file, fpath)

        self._unzip()

        self._maybe_to_netcdf('camels_aus_dyn')

    @property
    def start(self):
        return "19570101"

    @property
    def end(self):
        return "20181231"

    @property
    def location(self):
        return "Australia"

    def stations(self, as_list=True) -> list:
        fname = os.path.join(self.ds_dir, f"01_id_name_metadata{SEP}01_id_name_metadata{SEP}id_name_metadata.csv")
        df = pd.read_csv(fname)
        if as_list:
            return df['station_id'].to_list()
        else:
            return df

    @property
    def static_attribute_categories(self):
        attributes = []
        path = os.path.join(self.ds_dir, f'04_attributes{SEP}04_attributes')
        for f in os.listdir(path):
            if os.path.isfile(os.path.join(path, f)) and f.endswith('csv'):
                f = str(f.split('.csv')[0])
                attributes.append(''.join(f.split('_')[2:]))
        return attributes

    @property
    def static_features(self) -> list:
        static_fpath = os.path.join(self.ds_dir, 'static_features.csv')
        if not os.path.exists(static_fpath):
            files = glob.glob(f"{os.path.join(self.ds_dir, '04_attributes', '04_attributes')}/*.csv")
            cols = []
            for f in files:
                _df = pd.read_csv(f, index_col='station_id', nrows=1)
                cols += list(_df.columns)
        else:
            df = pd.read_csv(static_fpath, index_col='station_id', nrows=1)
            cols = list(df.columns)

        return cols

    @property
    def dynamic_features(self) -> list:
        return list(self.folders.keys())

    def _read_static(self, stations, attributes,
                     st=None, en=None):

        attributes = check_attributes(attributes, self.static_features)
        static_fname = 'static_features.csv'
        static_fpath = os.path.join(self.ds_dir, static_fname)
        if os.path.exists(static_fpath):
            static_df = pd.read_csv(static_fpath, index_col='station_id')
        else:
            files = glob.glob(f"{os.path.join(self.ds_dir, '04_attributes', '04_attributes')}/*.csv")
            static_df = pd.DataFrame()
            for f in files:
                _df = pd.read_csv(f, index_col='station_id')
                static_df = pd.concat([static_df, _df], axis=1)
            static_df.to_csv(static_fpath)

        static_df.index = static_df.index.astype(str)
        df = static_df.loc[stations][attributes]
        if isinstance(df, pd.Series):
            df = pd.DataFrame(df).transpose()

        return self.to_ts(df, st, en)

    def _read_dynamic_from_csv(self, stations, dynamic_features, **kwargs):

        dyn_attrs = {}
        dyn = {}
        for _attr in dynamic_features:

            _path = os.path.join(self.ds_dir, f'{self.folders[_attr]}.csv')
            _df = pd.read_csv(_path, na_values=['-99.99'])
            _df.index = pd.to_datetime(_df[['year', 'month', 'day']])
            [_df.pop(col) for col in ['year', 'month', 'day']]

            dyn_attrs[_attr] = _df

        # making one separate dataframe for one station
        for stn in stations:
            stn_df = pd.DataFrame()
            for attr, attr_df in dyn_attrs.items():
                if attr in dynamic_features:
                    stn_df[attr] = attr_df[stn]
            dyn[stn] = stn_df

        return dyn

    def fetch_static_features(self,
                              station,
                              features='all',
                              **kwargs) -> pd.DataFrame:
        """Fetches static attribuets of one station as dataframe."""

        return self._read_static(station, features)

    def plot(self, what, stations=None, **kwargs):
        assert what in ['outlets', 'boundaries']
        f1 = os.path.join(self.ds_dir, f'02_location_boundary_area{SEP}02_location_boundary_area{SEP}shp{SEP}CAMELS_AUS_BasinOutlets_adopted.shp')
        f2 = os.path.join(self.ds_dir, f'02_location_boundary_area{SEP}02_location_boundary_area{SEP}shp{SEP}bonus data{SEP}Australia_boundaries.shp')

        if plot_shapefile is not None:
            return plot_shapefile(f1, bbox_shp=f2, recs=stations, rec_idx=0, **kwargs)
        else:
            raise ModuleNotFoundError("Shapely must be installed in order to plot the datasets.")


class CAMELS_CL(Camels):
    """
    Downloads and processes CAMELS dataset of Chile
    https://doi.org/10.5194/hess-22-5817-2018
    """

    urls = {
        "1_CAMELScl_attributes.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "2_CAMELScl_streamflow_m3s.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "3_CAMELScl_streamflow_mm.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "4_CAMELScl_precip_cr2met.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "5_CAMELScl_precip_chirps.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "6_CAMELScl_precip_mswep.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "7_CAMELScl_precip_tmpa.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "8_CAMELScl_tmin_cr2met.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "9_CAMELScl_tmax_cr2met.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "10_CAMELScl_tmean_cr2met.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "11_CAMELScl_pet_8d_modis.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "12_CAMELScl_pet_hargreaves.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "13_CAMELScl_swe.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "14_CAMELScl_catch_hierarchy.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
        "CAMELScl_catchment_boundaries.zip": "https://store.pangaea.de/Publications/Alvarez-Garreton-etal_2018/",
    }

    dynamic_features = ['streamflow_m3s', 'streamflow_mm',
                        'precip_cr2met', 'precip_chirps', 'precip_mswep', 'precip_tmpa',
                        'tmin_cr2met', 'tmax_cr2met', 'tmean_cr2met',
                        'pet_8d_modis', 'pet_hargreaves',
                        'swe'
                        ]
    """
    Arguments:
        path: path where the CAMELS-AUS dataset has been downloaded. This path must
              contain five zip files and one xlsx file.
    """
    def __init__(self,
                 path: str = None
                 ):

        self.ds_dir = path

        super().__init__()

        if not os.path.exists(self.ds_dir):
            os.makedirs(self.ds_dir)

        for _file, url in self.urls.items():
            fpath = os.path.join(self.ds_dir, _file)
            if not os.path.exists(fpath):
                download(url+_file, fpath)
        self._unzip()

        self.dyn_fname = os.path.join(self.ds_dir, 'camels_cl_dyn.nc')
        self._maybe_to_netcdf('camels_cl_dyn')

    @property
    def _all_dirs(self):
        """All the folders in the dataset_directory"""
        return [f for f in os.listdir(self.ds_dir) if os.path.isdir(os.path.join(self.ds_dir, f))]

    @property
    def start(self):
        return "19130215"

    @property
    def end(self):
        return "20180309"

    @property
    def static_features(self) -> list:
        path = os.path.join(self.ds_dir, f"1_CAMELScl_attributes{SEP}1_CAMELScl_attributes.txt")
        df = pd.read_csv(path, sep='\t', index_col='gauge_id')
        return df.index.to_list()

    def stations(self) -> list:
        """Tells all station ids for which a data of a specific attribute is available."""
        stn_fname = os.path.join(self.ds_dir, 'stations.json')
        if not os.path.exists(stn_fname):
            _stations = {}
            for dyn_attr in self.dynamic_features:
                for _dir in self._all_dirs:
                    if dyn_attr in _dir:
                        fname = os.path.join(self.ds_dir, f"{_dir}{SEP}{_dir}.txt")
                        df = pd.read_csv(fname, sep='\t', nrows=2, index_col='gauge_id')
                        _stations[dyn_attr] = list(df.columns)

            stns = list(set.intersection(*map(set, list(_stations.values()))))
            with open(stn_fname, 'w') as fp:
                json.dump(stns, fp)
        else:
            with open(stn_fname, 'r') as fp:
                stns = json.load(fp)
        return stns

    def _read_dynamic_from_csv(self, stations, dynamic_features, st=None, en=None):

        dyn = {}
        st, en = self._check_length(st, en)

        assert all(stn in self.stations() for stn in stations)

        dynamic_features = check_attributes(dynamic_features, self.dynamic_features)

        # reading all dynnamic attributes
        dyn_attrs = {}
        for attr in dynamic_features:
            fname = [f for f in self._all_dirs if '_' + attr in f][0]
            fname = os.path.join(self.ds_dir, f'{fname}{SEP}{fname}.txt')
            _df = pd.read_csv(fname, sep='\t', index_col=['gauge_id'], na_values=" ")
            _df.index = pd.to_datetime(_df.index)
            dyn_attrs[attr] = _df[st:en]

        # making one separate dataframe for one station
        for stn in stations:
            stn_df = pd.DataFrame()
            for attr, attr_df in dyn_attrs.items():
                if attr in dynamic_features:
                    stn_df[attr] = attr_df[stn]
            dyn[stn] = stn_df[st:en]

        return dyn

    def _read_static(self, stations: list, attributes: list) -> pd.DataFrame:
        # overwritten for speed
        stns_df = pd.DataFrame(columns=attributes)
        path = os.path.join(self.ds_dir, f"1_CAMELScl_attributes{SEP}1_CAMELScl_attributes.txt")
        _df = pd.read_csv(path, sep='\t', index_col='gauge_id')

        for stn in stations:
            df = pd.DataFrame()
            if stn in _df:
                df[stn] = _df[stn]
            elif ' ' + stn in _df:
                df[stn] = _df[' ' + stn]

            stns_df = stns_df.append(df.transpose()[attributes])

        return stns_df

    def fetch_static_features(self,
                              station,
                              features=None,
                              st=None,
                              en=None
                              ):

        attributes = check_attributes(features, self.static_features)

        if isinstance(station, str):
            station = [station]

        return self._read_static(station, attributes)


class HYPE(Camels):
    """
    Downloads and preprocesses HYPE dataset from https://zenodo.org/record/4029572.
    This is a rainfall-runoff dataset of 564 stations from 1985 to 2019 at daily
    monthly and yearly time steps.
    paper : https://doi.org/10.2166/nh.2010.007
    """
    url = [
        "https://zenodo.org/record/581435",
        "https://zenodo.org/record/4029572"
    ]
    dynamic_features = [
        'AET_mm',
        'Baseflow_mm',
        'Infiltration_mm',
        'SM_mm',
        'Streamflow_mm',
        'Runoff_mm',
        'Qsim_m3-s',
        'Prec_mm',
        'PET_mm'
    ]

    def __init__(self, time_step: str = 'daily', **kwargs):
        assert time_step in ['daily', 'month', 'year']
        self.time_step = time_step
        self.ds_dir = None
        super().__init__(**kwargs)

        self._download()

        fpath = os.path.join(self.ds_dir, 'hype_year_dyn.nc')
        if not os.path.exists(fpath):
            self.time_step = 'daily'
            self._maybe_to_netcdf('hype_daily_dyn')
            self.time_step = 'month'
            self._maybe_to_netcdf('hype_month_dyn')
            self.time_step = 'year'
            self._maybe_to_netcdf('hype_year_dyn')
            self.time_step = time_step

        self.dyn_fname = os.path.join(self.ds_dir, f'hype_{time_step}_dyn.nc')

    def stations(self) -> list:
        _stations = np.arange(1, 565).astype(str)
        return list(_stations)

    @property
    def static_features(self):
        return []

    def _read_dynamic_from_csv(self,
                               stations: list,
                               attributes: Union[str, list] = 'all',
                               st=None,
                               en=None,
                               ):

        dynamic_features = check_attributes(attributes, self.dynamic_features)

        _dynamic_attributes = []
        for dyn_attr in dynamic_features:
            pref, suff = dyn_attr.split('_')[0], dyn_attr.split('_')[-1]
            _dyn_attr = f"{pref}_{self.time_step}_{suff}"
            _dynamic_attributes.append(_dyn_attr)

        df_attrs = {}
        for dyn_attr in _dynamic_attributes:
            fname = f"{dyn_attr}.csv"
            fpath = os.path.join(self.ds_dir, fname)
            index_col_name = 'DATE'
            if fname in ['SM_month_mm.csv', 'SM_year_mm.csv']:
                index_col_name = 'Date'
            _df = pd.read_csv(fpath, index_col=index_col_name)
            _df.index = pd.to_datetime(_df.index)
            # todo, some stations have wider range than self.st/self.en
            df_attrs[dyn_attr] = _df.loc[self.start:self.end]

        stns_dfs = {}
        for st in stations:
            stn_dfs = []
            cols = []
            for dyn_attr, dyn_df in df_attrs.items():
                stn_dfs.append(dyn_df[st])
                col_name = f"{dyn_attr.split('_')[0]}_{dyn_attr.split('_')[-1]}"  # get original name without time_step
                cols.append(col_name)
            stn_df = pd.concat(stn_dfs, axis=1)
            stn_df.columns = cols
            stns_dfs[st] = stn_df

        return stns_dfs

    def fetch_static_features(self, station, features):
        raise ValueError(f'No static feature for {self.name}')

    @property
    def start(self):
        return '19850101'

    @property
    def end(self):
        return '20191231'



import re
from weakref import WeakKeyDictionary

import numpy as np


metric_dict = {
    'Exa': 1e18,
    'Peta': 1e15,
    'Tera': 1e12,
    'Giga': 1e9,
    'Mega': 1e6,
    'Kilo': 1e3,
    'Hecto': 1e2,
    'Deca': 1e1,
    None: 1,
    'Deci':  1e-1,
    'Centi': 1e-2,
    'Milli': 1e-3,
    'Micro': 1e-6,
    'Nano': 1e-9,
    'Pico': 1e-12,
    'Femto': 1e-15,
    'Atto': 1e-18
}

time_dict = {
        'Year':   31540000,
        'Month':  2628000,
        'Weak':   604800,
        'Day':    86400,
        'Hour':   3600,
        'Minute': 60,
        'Second': 1,
    }

imperial_dist_dict = {
        'Mile': 63360,
        'Furlong': 7920,
        'Rod': 198,
        'Yard': 36,
        'Foot': 12,
        'Inch': 1
    }


unit_plurals = {
    "Inches": "Inch",
    "Miles": "Mile",
    "Meters": "Meter",
    "Feet": "Foot"
}


def check_plurals(unit):
    if unit in unit_plurals:
        unit = unit_plurals[unit]
    return unit


def split_speed_units(unit):
    dist = unit.split("Per")[0]
    zeit = unit.split("Per")[1]

    # distance and speed may contain underscore ('_') at start or at end e.g. when unit is "Meter_Per_Second"
    # we need to remove all such underscores
    dist = dist.replace("_", "")
    zeit = zeit.replace("_", "")
    if dist in unit_plurals:
        dist = unit_plurals[dist]
    return dist, zeit


class WrongUnitError(Exception):
    def __init__(self, u_type, qty, unit, allowed, prefix=None):
        self.u_type = u_type
        self.qty = qty
        self.unit = unit
        self.allowed = allowed
        self.pre = prefix

    def __str__(self):
        if self.pre is None:
            return '''
*
*   {} unit `{}` for {} is wrong. Use either of {}
*
'''.format(self.u_type, self.unit, self.qty, self.allowed)
# prefix {milli} provided for {input} unit of {temperature} is wrong. {input} unit is {millipascal}, allowed are {}}
        else:
            return """
*
* prefix `{}` provided for {} unit of {} is wrong.
* {} unit is: {}. Allowed units are
* {}.
*
""".format(self.pre, self.u_type, self.qty, self.u_type, self.unit, self.allowed)


def check_converter(converter):
    super_keys = converter.keys()

    for k, v in converter.items():
        sub_keys = v.keys()

        if all(x in super_keys for x in sub_keys):
            a = 1
        else:
            a = 0

        if all(x in sub_keys for x in super_keys):
            b = 1
        else:
            b = 0

        assert a == b


TempUnitConverter = {
    "FAHRENHEIT": {
        "Fahrenheit": lambda fahrenheit: fahrenheit * 1.0,  # fahrenheit to Centigrade
        "Kelvin": lambda fahrenheit: (fahrenheit + 459.67) * 5/9,  # fahrenheit to kelvin
        "Centigrade": lambda fahrenheit: (fahrenheit - 32.0) / 1.8  # fahrenheit to Centigrade
    },
    "KELVIN": {
        "Fahrenheit": lambda kelvin: kelvin * 9/5 - 459.67,  # kelvin to fahrenheit
        "Kelvin": lambda k: k*1.0,     # Kelvin to Kelvin
        "Centigrade": lambda kelvin: kelvin - 273.15  # kelvin to Centigrade}
    },
    "CENTIGRADE": {
        "Fahrenheit": lambda centigrade: centigrade * 1.8 + 32,  # Centigrade to fahrenheit
        "Kelvin": lambda centigrade: centigrade + 273.15,  # Centigrade to kelvin
        "Centigrade": lambda centigrade: centigrade * 1.0}
}


PressureConverter = {
    "Pascal": {  # Pascal to
        "Pascal": lambda pascal: pascal,
        "Bar": lambda pascal: pascal * 1e-5,
        "Atm": lambda pascal: pascal / 101325,
        "Torr": lambda pascal: pascal * 0.00750062,
        "Psi": lambda pascal: pascal / 6894.76,
        "Ta": lambda pascal: pascal * 1.01971621298E-5
    },
    "Bar": {  # Bar to
        "Pascal": lambda bar: bar / 0.00001,
        "Bar": lambda bar: bar,
        "Atm": lambda bar: bar / 1.01325,
        "Torr": lambda bar: bar * 750.062,
        "Psi": lambda bar: bar * 14.503,
        "Ta": lambda bar: bar * 1.01972
    },
    "Atm": {  # Atm to
        "Pascal": lambda atm: atm * 101325,
        "Bar": lambda atm: atm * 1.01325,
        "Atm": lambda atm: atm,
        "Torr": lambda atm: atm * 760,
        "Psi": lambda atm: atm * 14.6959,
        "At": lambda atm: atm * 1.03322755477
    },
    "Torr": {  # Torr to
        "Pascal": lambda torr: torr / 0.00750062,
        "Bar": lambda torr: torr / 750.062,
        "Atm": lambda torr: torr / 760,
        "Torr": lambda tor: tor,
        "Psi": lambda torr: torr / 51.7149,
        "Ta": lambda torr: torr * 0.00135950982242
    },
    "Psi": {  # Psi to
        "Pascal": lambda psi: psi * 6894.76,
        "Bar": lambda psi: psi / 14.5038,
        "Atm": lambda psi: psi / 14.6959,
        "Torr": lambda psi: psi * 51.7149,
        "Psi": lambda psi: psi,
        "Ta": lambda psi: psi * 0.0703069578296,
    },
    "Ta": {   # Ta to
        "Pascal": lambda at: at / 1.01971621298E-5,
        "Bar": lambda at: at / 1.0197,
        "Atm": lambda at: at / 1.03322755477,
        "Torr": lambda at: at / 0.00135950982242,
        "Psi": lambda at: at / 0.0703069578296,
        "Ta": lambda ta: ta
    }
}

DistanceConverter = {
    "Meter": {
        "Meter": lambda meter: meter,
        "Inch": lambda meter: meter * 39.3701
    },
    "Inch": {
        "Meter": lambda inch: inch * 0.0254,
        "Inch": lambda inch: inch
    }
}


class Pressure(object):
    """
    ```python
    p = Pressure(20, "Pascal")
    print(p.MilliBar)    #>> 0.2
    print(p.Bar)         #>> 0.0002
    p = Pressure(np.array([10, 20]), "KiloPascal")
    print(p.MilliBar)    # >> [100, 200]
    p = Pressure(np.array([1000, 2000]), "MilliBar")
    print(p.KiloPascal)  #>> [100, 200]
    print(p.Atm)         # >> [0.98692, 1.9738]
    ```
    """

    def __init__(self, val, input_unit):
        self.val = val
        check_converter(PressureConverter)
        self.input_unit = input_unit

    @property
    def allowed(self):
        return list(PressureConverter.keys())

    @property
    def input_unit(self):
        return self._input_unit

    @input_unit.setter
    def input_unit(self, in_unit):

        self._input_unit = in_unit

    def __getattr__(self, out_unit):
        # pycharm calls this method for its own working, executing default behaviour at such calls
        if out_unit.startswith('_'):
            return self.__getattribute__(out_unit)
        else:
            act_iu, iu_pf = self._preprocess(self.input_unit, "Input")

            act_ou, ou_pf = self._preprocess(out_unit, "Output")

            if act_iu not in self.allowed:
                raise WrongUnitError("Input", self.__class__.__name__, act_iu, self.allowed)
            if act_ou not in self.allowed:
                raise WrongUnitError("output", self.__class__.__name__, act_ou, self.allowed)

            ou_f = PressureConverter[act_iu][act_ou](self.val)

            val = np.round(np.array((iu_pf * ou_f) / ou_pf), 5)
            return val

    def _preprocess(self, given_unit, io_type="Input"):
        split_u = split_units(given_unit)
        if len(split_u) < 1:  # Given unit contained no capital letter so list is empty
            raise WrongUnitError(io_type, self.__class__.__name__, given_unit, self.allowed)

        pf, ou_pf = 1.0, 1.0
        act_u = split_u[0]
        if len(split_u) > 1:
            pre_u = split_u[0]  # prefix of input unit
            act_u = split_u[1]  # actual input unit

            if pre_u in metric_dict:
                pf = metric_dict[pre_u]  # input unit prefix factor
            else:
                raise WrongUnitError(io_type, self.__class__.__name__, act_u, self.allowed, pre_u)

        return act_u, pf


class NotString:
    def __init__(self):
        self.data = WeakKeyDictionary()

    def __get__(self, instance, owner):
        return self.data[instance]

    def __set__(self, instance, value):
        if isinstance(value, list) or isinstance(value, np.ndarray):
            value = np.array(value).astype(np.float32)

        self.data[instance] = value

    def __set_name__(self, owner, name):
        self.name = name


class Temp(object):
    """
    The idea is to write the conversion functions in a dictionary and then dynamically create attribute it the attribute
    is present in converter as key otherwise raise WongUnitError.
    converts temperature among units [kelvin, centigrade, fahrenheit]
    :param `temp`  a numpy array
    :param `input_unit` str, units of temp, should be "Kelvin", "Centigrade" or "Fahrenheit"

    Example:
    ```python
    temp = np.arange(10)
    T = Temp(temp, 'Centigrade')
    T.Kelvin
    >> array([273 274 275 276 277 278 279 280 281 282])
    T.Fahrenheit
    >> array([32. , 33.8, 35.6, 37.4, 39.2, 41. , 42.8, 44.6, 46.4, 48.2])
    T.Centigrade
    >>array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    ```
    """
    val = NotString()

    def __init__(self, val, input_unit):
        self.val = val
        check_converter(TempUnitConverter)
        self.input_unit = input_unit

    def __getattr__(self, out_unit):
        # pycharm calls this method for its own working, executing default behaviour at such calls
        if out_unit.startswith('_'):
            return self.__getattribute__(out_unit)
        else:
            if out_unit not in TempUnitConverter[self.input_unit]:
                raise WrongUnitError("output", self.__class__.__name__, out_unit, self.allowed)

            val = TempUnitConverter[self.input_unit][str(out_unit)](self.val)
            return val

    @property
    def allowed(self):
        return list(TempUnitConverter.keys())

    @property
    def input_unit(self):
        return self._input_unit

    @input_unit.setter
    def input_unit(self, in_unit):
        if in_unit.upper() == 'CELCIUS':
            in_unit = 'CENTIGRADE'
        if in_unit.upper() not in self.allowed:
            raise WrongUnitError("Input", self.__class__.__name__, in_unit, self.allowed)
        self._input_unit = in_unit.upper()


class Distance(object):
    """
    unit converter for distance or length between different imperial and/or metric units.
    ```python
    t = Distance(np.array([2.0]), "Mile")
    np.testing.assert_array_almost_equal(t.Inch, [126720], 5)
    np.testing.assert_array_almost_equal(t.Meter, [3218.688], 5)
    np.testing.assert_array_almost_equal(t.KiloMeter, [3.218688], 5)
    np.testing.assert_array_almost_equal(t.CentiMeter, [321869], 0)
    np.testing.assert_array_almost_equal(t.Foot, [10560.], 5)

    t = Distance(np.array([5000]), "MilliMeter")
    np.testing.assert_array_almost_equal(t.Inch, [196.85039], 5)
    np.testing.assert_array_almost_equal(t.Meter, [5.0], 5)
    np.testing.assert_array_almost_equal(t.KiloMeter, [0.005], 5)
    np.testing.assert_array_almost_equal(t.CentiMeter, [500.0], 5)
    np.testing.assert_array_almost_equal(t.Foot, [16.404199], 5)
    ```
    """

    def __init__(self, val, input_unit):
        self.val = val
        self.input_unit = input_unit

    @property
    def allowed(self):
        return list(imperial_dist_dict.keys()) + ['Meter']

    @property
    def input_unit(self):
        return self._input_unit

    @input_unit.setter
    def input_unit(self, in_unit):
        self._input_unit = in_unit

    def __getattr__(self, out_unit):
        # pycharm calls this method for its own working, executing default behaviour at such calls
        if out_unit.startswith('_'):
            return self.__getattribute__(out_unit)
        else:
            act_iu, iu_pf = self._preprocess(self.input_unit, "Input")

            act_ou, ou_pf = self._preprocess(out_unit, "Output")

            act_iu = check_plurals(act_iu)
            act_ou = check_plurals(act_ou)

            if act_iu not in self.allowed:
                raise WrongUnitError("Input", self.__class__.__name__, act_iu, self.allowed)
            if act_ou not in self.allowed:
                raise WrongUnitError("output", self.__class__.__name__, act_ou, self.allowed)

            out_in_meter = self._to_meters(ou_pf, act_ou)  # get number of meters in output unit
            input_in_meter = self.val * iu_pf  # for default case when input unit has Meter in it

            # if input unit is in imperial system, first convert it into inches and then into meters
            if act_iu in imperial_dist_dict:
                input_in_inches = imperial_dist_dict[act_iu] * self.val * iu_pf
                input_in_meter = DistanceConverter['Inch']['Meter'](input_in_inches)

            val = input_in_meter / out_in_meter

            return val

    def _to_meters(self, prefix, actual_unit):
        meters = prefix
        if actual_unit != "Meter":
            inches = imperial_dist_dict[actual_unit] * prefix
            meters = DistanceConverter['Inch']['Meter'](inches)
        return meters

    def _preprocess(self, given_unit, io_type="Input"):
        # TODO unit must not be split based on capital letters, it is confusing and prone to erros
        split_u = split_units(given_unit)
        if len(split_u) < 1:  # Given unit contained no capital letter so list is empty
            raise WrongUnitError(io_type, self.__class__.__name__, given_unit, self.allowed)

        pf, ou_pf = 1.0, 1.0
        act_u = split_u[0]
        if len(split_u) > 1:
            pre_u = split_u[0]  # prefix of input unit
            act_u = split_u[1]  # actual input unit

            if pre_u in metric_dict:
                pf = metric_dict[pre_u]  # input unit prefix factor
            else:
                raise WrongUnitError(io_type, self.__class__.__name__, act_u, self.allowed, pre_u)

        return act_u, pf


class Time(object):

    """
    ```python
    t = Time(np.array([100, 200]), "Hour")
    np.testing.assert_array_almost_equal(t.Day, [4.16666667, 8.33333333], 5)
    t = Time(np.array([48, 24]), "Day")
    np.testing.assert_array_almost_equal(t.Minute, [69120., 34560.], 5)
    ```
    """
    def __init__(self, val, input_unit):
        self.val = val
        self.input_unit = input_unit

    @property
    def allowed(self):
        return list(time_dict.keys())

    @property
    def input_unit(self):
        return self._input_unit

    @input_unit.setter
    def input_unit(self, in_unit):
        self._input_unit = in_unit

    def __getattr__(self, out_unit):
        # pycharm calls this method for its own working, executing default behaviour at such calls
        if out_unit.startswith('_'):
            return self.__getattribute__(out_unit)
        else:
            act_iu, iu_pf = self._preprocess(self.input_unit, "Input")

            act_ou, ou_pf = self._preprocess(out_unit, "Output")

            if act_iu not in self.allowed:
                raise WrongUnitError("Input", self.__class__.__name__, act_iu, self.allowed)
            if act_ou not in self.allowed:
                raise WrongUnitError("output", self.__class__.__name__, act_ou, self.allowed)

            in_sec = time_dict[act_iu] * self.val * iu_pf
            val = in_sec / (time_dict[act_ou]*ou_pf)
            return val

    def _preprocess(self, given_unit, io_type="Input"):
        split_u = split_units(given_unit)
        if len(split_u) < 1:  # Given unit contained no capital letter so list is empty
            raise WrongUnitError(io_type, self.__class__.__name__, given_unit, self.allowed)

        pf, ou_pf = 1.0, 1.0
        act_u = split_u[0]
        if len(split_u) > 1:
            pre_u = split_u[0]  # prefix of input unit
            act_u = split_u[1]  # actual input unit

            if pre_u in metric_dict:
                pf = metric_dict[pre_u]  # input unit prefix factor
            else:
                raise WrongUnitError(io_type, self.__class__.__name__, act_u, self.allowed, pre_u)

        return act_u, pf


class Speed(object):
    """
    converts between different units using Distance and Time classes which convert
    distance and time units separately. This class both classes separately and
    then does the rest of the work.
    ```python
    s = Speed(np.array([10]), "KiloMeterPerHour")
    np.testing.assert_array_almost_equal(s.MeterPerSecond, [2.77777778], 5)
    np.testing.assert_array_almost_equal(s.MilePerHour, [6.21371192], 5)
    np.testing.assert_array_almost_equal(s.FootPerSecond, [9.11344415], 5)
    s = Speed(np.array([14]), "FootPerSecond")
    np.testing.assert_array_almost_equal(s.MeterPerSecond, [4.2672], 5)
    np.testing.assert_array_almost_equal(s.MilePerHour, [9.54545], 5)
    np.testing.assert_array_almost_equal(s.KiloMeterPerHour, [15.3619], 4)
    s = Speed(np.arange(10), 'KiloMetersPerHour')
    o = np.array([ 0. , 10.936, 21.872, 32.808, 43.744, 54.680, 65.616 , 76.552, 87.489, 98.425])
    np.testing.assert_array_almost_equal(s.InchesPerSecond, o, 2)
    ```
    """

    def __init__(self, val, input_unit):
        self.val = val
        self.input_unit = input_unit

    @property
    def input_unit(self):
        return self._input_unit

    @input_unit.setter
    def input_unit(self, in_unit):
        if '_' in in_unit:
            raise ValueError("remove underscore from units {}".format(in_unit))
        self._input_unit = in_unit

    def __getattr__(self, out_unit):
        # pycharm calls this method for its own working, executing default behaviour at such calls
        if out_unit.startswith('_'):
            return self.__getattribute__(out_unit)
        else:
            in_dist, in_zeit = split_speed_units(self.input_unit)
            out_dist, out_zeit = split_speed_units(out_unit)

            d = Distance(np.array([1]), in_dist)
            dist_f = getattr(d, out_dist)  # distance factor
            t = Time(np.array([1]), in_zeit)
            time_f = getattr(t, out_zeit)   # time factor

            out_val = self.val * (dist_f/time_f)
            return out_val


def split_units(unit):
    """splits string `unit` based on capital letters"""
    return re.findall('[A-Z][^A-Z]*', unit)




from torch import nn

class Conditionalize(nn.Module):

    def __init__(self, input_size, output_size, num_max_conditions=10, **kwargs):

        super(Conditionalize, self).__init__()

        self.units = output_size

        self.cond_to_init_state_dense_1 = nn.Linear(in_features=input_size, out_features=output_size)


    def forward(self, inputs):
        cond = inputs
        cond_state = self.cond_to_init_state_dense_1(cond)

        return cond_state


i = nn.LSTM(3)



import os
from typing import Union, Callable, List

try:
    import shap
except ModuleNotFoundError:
    shap = None

import scipy as sp
import numpy as np
import pandas as pd
from easy_mpl import imshow
import matplotlib.pyplot as plt

try:
    import tensorflow.keras.backend as K
except ModuleNotFoundError:
    K = None

from ai4water.backend import sklearn_models
from ._explain import ExplainerMixin
from .utils import convert_ai4water_model


class ShapExplainer(ExplainerMixin):
    """
    Wrapper around SHAP `explainers` and `plots` to draw and save all the plots
    for a given model.

    Attributes:
        features :
        train_summary : only for KernelExplainer
        explainer :
        shap_values :


    Methods
    --------
    - summary_plot
    - force_plot_single_example
    - dependence_plot_single_feature
    - force_plot_all

    Examples:
        >>> from ai4water.postprocessing.explain import ShapExplainer
        >>> from sklearn.model_selection import train_test_split
        >>> from sklearn import linear_model
        >>> import shap
        ...
        >>> X,y = shap.datasets.diabetes()
        >>> X_train,X_test,y_train,y_test = train_test_split(X, y, test_size=0.2, random_state=0)
        >>> lin_regr = linear_model.LinearRegression()
        >>> lin_regr.fit(X_train, y_train)
        >>> explainer = ShapExplainer(lin_regr, X_test, X_train, num_means=10)
        >>> explainer()

    """

    allowed_explainers = [
        "Explainer",
        "DeepExplainer",
        "TreeExplainer",
        "KernelExplainer",
        "LinearExplainer",
        "AdditiveExplainer",
        "GPUTreeExplainer",
        "GradientExplainer",
        "PermutationExplainer",
        "SamplingExplainer",
        "PartitionExplainer"
    ]

    def __init__(
            self,
            model,
            data: Union[np.ndarray, pd.DataFrame, List[np.ndarray]],
            train_data: Union[np.ndarray, pd.DataFrame, List[np.ndarray]] = None,
            explainer: Union[str, Callable] = None,
            num_means: int = 10,
            path: str = None,
            feature_names: list = None,
            framework: str = None,
            layer: Union[int, str] = None,
    ):
        """

        Args:
            model :
                a Model/regressor/classifier from sklearn/xgboost/catboost/LightGBM/tensorflow/pytorch/ai4water
                The model must have a `predict` method.
            data :
                Data on which to make interpretation. Its dimension should be
                same as that of training data. It can be either training or test
                data
            train_data :
                The data on which the `model` was trained. It is used to
                get train_summary. It can a numpy array or a pandas DataFrame.
                Only required for scikit-learn based models.
            explainer : str
                the explainer to use. If not given, the explainer will be inferred.
            num_means : int
                Numher of means, used in `shap.kmeans` to calculate train_summary
                using shap.kmeans. Only used when explainer is "KernelExplainer"
            path : str
                path to save the plots. By default, plots will be saved in current
                working directory
            feature_names : list
                Names of features. Should only be given if train/test data is numpy
                array.
            framework : str
                either "DL" or "ML". Here "DL" shows that the `model` is a deep
                learning or neural network based model and "ML" represents other
                models. For "DL" the explainer will be either "DeepExplainer" or
                "GradientExplainer". If not given, it will be inferred. In such
                a case "DeepExplainer" will be prioritized over "GradientExplainer"
                for DL frameworks and "TreeExplainer" will be prioritized for "ML"
                frameworks.
            layer : Union[int, str]
                only relevant when framework is "DL" i.e when the model consits of layers
                of neural networks.

        """
        test_data = maybe_to_dataframe(data, feature_names)
        train_data = maybe_to_dataframe(train_data, feature_names)

        super(ShapExplainer, self).__init__(path=path or os.getcwd(), data=test_data, features=feature_names)

        if train_data is None:
            self._check_data(test_data)
        else:
            self._check_data(train_data, test_data)

        model, framework, explainer, model_name = convert_ai4water_model(model,
                                                                         framework,
                                                                         explainer)
        self.is_sklearn = True
        if model_name not in sklearn_models:
            if model_name in ["XGBRegressor",
                              "XGBClassifier",
                              "LGBMRegressor",
                              "LGBMClassifier",
                              "CatBoostRegressor",
                              "CatBoostClassifier"  
                              "XGBRFRegressor"
                              "XGBRFClassifier"
                                            ]:
                self.is_sklearn = False

            elif not self._is_dl(model):
                raise ValueError(f"{model.__class__.__name__} is not a valid model model")

        self._framework = self.infer_framework(model, framework, layer, explainer)

        self.model = model
        self.data = test_data
        self.layer = layer
        self.features = feature_names

        self.explainer = self._get_explainer(explainer, train_data=train_data, num_means=num_means)

        self.shap_values = self.get_shap_values(test_data)

    @staticmethod
    def _is_dl(model):
        if hasattr(model, "count_params") or hasattr(model, "named_parameters"):
            return True
        return False

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, x):
        if x is not None:

            if not isinstance(x, str):
                assert isinstance(x, int), f"layer must either b string or integer"
                assert x <= len(self.model.layers)  # todo, what about pytorch

        self._layer = x

    def map2layer(self, x, layer):
        feed_dict = dict(zip([self.model.layers[0].input], [x.copy()]))
        import tensorflow as tf
        if int(tf.__version__[0]) < 2:
            sess = K.get_session()
        else:
            sess = tf.compat.v1.keras.backend.get_session()

        if isinstance(layer, int):
            return sess.run(self.model.layers[layer].input, feed_dict)
        else:
            return sess.run(self.model.get_layer(layer).input, feed_dict)

    def infer_framework(self, model, framework, layer, explainer):
        if framework is not None:
            inf_framework = framework
        elif self._is_dl(model):
            inf_framework = "DL"
        elif isinstance(explainer, str) and explainer in ("DeepExplainer", "GradientExplainer"):
            inf_framework = "DL"
        elif explainer.__class__.__name__ in ("DeepExplainer", "GradientExplainer"):
            inf_framework = "DL"
        else:
            inf_framework = "ML"

        assert inf_framework in ("ML", "DL")

        if inf_framework != "DL":
            assert layer is None

        if inf_framework == "DL" and isinstance(explainer, str):
            assert explainer in ("DeepExplainer",
                                 "GradientExplainer",
                                 "PermutationExplainer"), f"invalid explainer {inf_framework}"

        return inf_framework

    def _get_explainer(self,
                       explainer: Union[str, Callable],
                       num_means,
                       train_data
                       ):

        if explainer is not None:
            if callable(explainer):
                return explainer

            assert isinstance(explainer, str), f"explainer should be callable or string but" \
                                               f" it is {explainer.__class__.__name__}"
            assert explainer in self.allowed_explainers, f"{explainer} is not a valid explainer"

            if explainer == "KernelExplainer":
                explainer = self._get_kernel_explainer(train_data, num_means)

            elif explainer == "DeepExplainer":
                explainer = self._get_deep_explainer()

            elif explainer == "GradientExplainer":
                explainer = self._get_gradient_explainer()

            elif explainer == "PermutationExplainer":
                explainer = shap.PermutationExplainer(self.model, self.data)

            else:
                explainer = getattr(shap, explainer)(self.model)

        else:  # explainer is not given explicitly, we need to infer it
            explainer = self._infer_explainer_to_use(train_data, num_means)

        return explainer

    def _get_kernel_explainer(self, data, num_means):

        assert isinstance(num_means,
                          int), f'num_means should be integer but given value is of type {num_means.__class__.__name__}'

        if data is None:
            raise ValueError("Provide train_data in order to use KernelExplainer.")

        self.train_summary = shap.kmeans(data, num_means)
        explainer = shap.KernelExplainer(self.model.predict, self.train_summary)

        return explainer

    def _infer_explainer_to_use(self, train_data, num_means):
        """Tries to infer explainer to use from the type of model."""
        # todo, Fig 3 of Lundberberg's Nature MI paper shows that TreeExplainer
        # performs better than KernelExplainer, so try to use supports_model_with_masker

        if self.model.__class__.__name__ in ["XGBRegressor", "LGBMRegressor", "CatBoostRegressor",
                                             "XGBRFRegressor"]:
            explainer = shap.TreeExplainer(self.model)

        elif self.model.__class__.__name__ in sklearn_models:
            explainer = self._get_kernel_explainer(train_data, num_means)

        elif self._framework == "DL":
            explainer = self._get_deep_explainer()
        else:
            raise ValueError(f"Can not infer explainer for model {self.model.__class__.__name__}."
                             f" Plesae specify explainer by using `explainer` keyword argument")
        return explainer

    def _get_deep_explainer(self):
        data = self.data.values if isinstance(self.data, pd.DataFrame) else self.data
        return getattr(shap, "DeepExplainer")(self.model, data)

    def _get_gradient_explainer(self):

        if self.layer is None:
            # GradientExplainer is also possible without specifying a layer
            return shap.GradientExplainer(self.model, self.data)
        if isinstance(self.layer, int):
            return shap.GradientExplainer((self.model.layers[self.layer].input, self.model.layers[-1].output),
                                          self.map2layer(self.data, self.layer))
        else:
            return shap.GradientExplainer((self.model.get_layer(self.layer).input, self.model.layers[-1].output),
                                          self.map2layer(self.data, self.layer))

    def _check_data(self, *data):
        if self.single_source:
            for d in data:
                assert type(d) == np.ndarray or type(d) == pd.DataFrame, f"""
                data must be numpy array or pandas dataframe but it is of type {d.__class__.__name__}"""

            assert len(set([d.ndim for d in data])) == 1, "train and test data should have same ndim"
            assert len(set([d.shape[-1] for d in data])) == 1, "train and test data should have same input features"
            assert len(set([type(d) for d in data])) == 1, "train and test data should be of same type"

        return

    def get_shap_values(self, data, **kwargs):

        if self.explainer.__class__.__name__ in ["Permutation"]:
            return self.explainer(data)

        elif self._framework == "DL":
            return self._shap_values_dl(data, **kwargs)

        return self.explainer.shap_values(data)

    def _shap_values_dl(self, data, ranked_outputs=None, **kwargs):
        """Gets the SHAP values"""
        data = data.values if isinstance(data, pd.DataFrame) else data

        if self.explainer.__class__.__name__ == "Deep":
            shap_values = self.explainer.shap_values(data, ranked_outputs=ranked_outputs, **kwargs)

        elif isinstance(self.explainer, shap.GradientExplainer) and self.layer is None:
            shap_values = self.explainer.shap_values(data, ranked_outputs=ranked_outputs, **kwargs)

        else:
            shap_values = self.explainer.shap_values(self.map2layer(data, self.layer),
                                                     ranked_outputs=ranked_outputs, **kwargs)
            if ranked_outputs:
                shap_values, indexes = shap_values

        return shap_values

    def __call__(self,
                 force_plots=True,
                 plot_force_all=False,
                 dependence_plots=False,
                 beeswarm_plots=False,
                 heatmap=False,
                 show=False,
                 save=True
                 ):
        """Draws and saves all the plots for a given sklearn model in the path.

        plot_force_all is set to False by default because it is causing
        Process finished error due. Avoiding this error is a complex function
        of scipy and numba versions.
        """

        if dependence_plots:
            for feature in self.features:
                self.dependence_plot_single_feature(feature, f"dependence_plot_{feature}", show=show, save=save)

        if force_plots:
            for i in range(self.data.shape[0]):

                self.force_plot_single_example(i, f"force_plot_{i}", show=show, save=save)

        if beeswarm_plots:
            self.beeswarm_plot(show=show, save=save)

        if plot_force_all:
            self.force_plot_all("force_plot_all", show=show, save=save)

        if heatmap:
            self.heatmap(show=show, save=save)

        self.summary_plot("summary_plot", show=show, save=save)

        return

    def summary_plot(
            self,
            name: str = "summary_plot",
            show: bool = True,
            save: bool = False,
            **kwargs
    ):
        """Plots the summary plot of SHAP package.

        Arguments:
            name:
                name of saved file
            show:
                whether to show the plot or not
            save:
                whether to save the plot or not
            kwargs:
                any keyword arguments to shap.summary_plot
        """

        def _summary_plot(_shap_val, _data, _features, _name):
            plt.close('all')
            shap.summary_plot(_shap_val, _data, show=False, feature_names=_features, **kwargs)
            if save:
                plt.savefig(os.path.join(self.path, _name), dpi=300, bbox_inches="tight")
            if show:
                plt.show()

            shap.summary_plot(_shap_val, _data, show=False, plot_type="bar", feature_names=_features,
                              **kwargs)
            if save:
                plt.savefig(os.path.join(self.path, _name + " _bar"), dpi=300, bbox_inches="tight")
            if show:
                plt.show()

            return

        shap_vals = self.shap_values
        if isinstance(shap_vals, list) and len(shap_vals) == 1:
            shap_vals = shap_vals[0]

        data = self.data

        if self.single_source:
            if data.ndim == 3:
                assert shap_vals.ndim == 3

                for lookback in range(data.shape[1]):

                    _summary_plot(shap_vals[:, lookback], data[:, lookback], self.features,  _name=f"{name}_{lookback}")
            else:
                _summary_plot(shap_vals, data, self.features, name)
        else:
            # data is a list of data sources
            for idx, _data in enumerate(data):
                if _data.ndim == 3:
                    for lb in range(_data.shape[1]):
                        _summary_plot(shap_vals[idx][:, lb], _data[:, lb], self.features[idx],
                                      _name=f"{name}_{idx}_{lb}")
                else:
                    _summary_plot(shap_vals[idx], _data, self.features[idx], _name=f"{name}_{idx}")

        return

    def force_plot_single_example(
            self,
            idx:int,
            name=None,
            show=True,
            save=False,
            **force_kws
    ):
        """Draws force_plot_
        for a single example/row/sample/instance/data point.

        If the data is 3d and shap values are 3d then they are unrolled/flattened
        before plotting

        Arguments:
            idx:
                index of exmaple to use. It can be any value >=0
            name:
                name of saved file
            show:
                whether to show the plot or not
            save:
                whether to save the plot or not
            force_kws : any keyword argument for force plot

        Returns:
            plotter object

        .. _force_plot:
            https://shap.readthedocs.io/en/latest/generated/shap.plots.force.html
        """

        shap_vals = self.shap_values

        if isinstance(shap_vals, list) and len(shap_vals) == 1:
            shap_vals = shap_vals[0]

        shap_vals = shap_vals[idx]

        if type(self.data) == np.ndarray:
            data = self.data[idx]
        else:
            data = self.data.iloc[idx, :]

        if self.explainer.__class__.__name__ == "Gradient":
            expected_value = [0]
        else:
            expected_value = self.explainer.expected_value

        features = self.features
        if data.ndim == 2 and shap_vals.ndim == 2:  # input was 3d i.e. ml model uses 3d input
            features = self.unrolled_features
            expected_value = expected_value[0]  # todo
            shap_vals = shap_vals.reshape(-1,)
            data = data.reshape(-1, )

        plt.close('all')

        plotter = shap.force_plot(
            expected_value,
            shap_vals,
            data,
            feature_names=features,
            show=False,
            matplotlib=True,
            **force_kws
        )

        if save:
            name = name or f"force_plot_{idx}"
            plotter.savefig(os.path.join(self.path, name), dpi=300, bbox_inches="tight")

        if show:
            plotter.show()

        return plotter

    def dependence_plot_all_features(self, show=True, save=False, **dependence_kws):
        """dependence plot for all features"""
        for feature in self.features:
            self.dependence_plot_single_feature(feature, f"dependence_plot_{feature}", show=show, save=save,
                                                **dependence_kws)
        return

    def dependence_plot_single_feature(self, feature, name="dependence_plot", show=False, save=True, **kwargs):
        """dependence_ plot for a single feature. See this_ .

        .. _dependence:
            https://slundberg.github.io/shap/notebooks/plots/dependence_plot.html
        .. _this:
            https://shap-lrjball.readthedocs.io/en/docs_update/generated/shap.dependence_plot.html
        """
        plt.close('all')

        if len(name) > 150:  # matplotlib raises error if the length of filename is too large
            name = name[0:150]

        shap_values = self.shap_values
        if isinstance(shap_values, list) and len(shap_values) == 1:
            shap_values = shap_values[0]

        shap.dependence_plot(feature,
                             shap_values,
                             self.data,
                             show=False,
                             **kwargs)
        if save:
            plt.savefig(os.path.join(self.path, name), dpi=300, bbox_inches="tight")
        if show:
            plt.show()
        return

    def force_plot_all(self, name="force_plot.html", save=True, show=True, **force_kws):
        """draws force plot for all examples in the given data and saves it in an html"""

        # following scipy versions cause kernel stoppage when calculating
        if sp.__version__ in ["1.4.1", "1.5.2", "1.7.1"]:
            print(f"force plot can not be plotted for scipy version {sp.__version__}. Please change your scipy")
            return

        shap_values = self.shap_values
        if isinstance(shap_values, list) and len(shap_values) == 1:
            shap_values = shap_values[0]
        plt.close('all')
        plot = shap.force_plot(self.explainer.expected_value, shap_values, self.data, **force_kws)
        if save:
            shap.save_html(os.path.join(self.path, name), plot)

        return

    def waterfall_plot_all_examples(
            self,
            name: str = "waterfall",
            save=True,
            show=False,
            **waterfall_kws
    ):
        """Plots the waterfall_ plot of SHAP package

        It plots for all the examples/instances from test_data.

        .. _waterfall:
            https://shap.readthedocs.io/en/latest/generated/shap.plots.waterfall.html
        """
        for i in range(len(self.data)):
            self.waterfall_plot_single_example(i, name=name, save=save, show=show, **waterfall_kws)

        return

    def waterfall_plot_single_example(
            self,
            example_index: int,
            name: str = "waterfall",
            show: bool = False,
            save=True,
            max_display: int = 10,
    ):
        """draws and saves waterfall_ plot
         for one example.

        The waterfall plots are based upon SHAP values and show the
        contribution by each feature in model's prediction. It shows which
        feature pushed the prediction in which direction. They answer the
        question, why the ML model simply did not predict mean of training y
        instead of what it predicted. The mean of training observations that
        the ML model saw during training is called base value or expected value.

        Arguments:
            example_index : int
                index of example to use
            max_display : int
                maximu features to display
            name : str
                name of plot
            save : bool
                whether to save the plot or not
            show : bool
                whether to show the plot or now

        .. _waterfall:
            https://shap.readthedocs.io/en/latest/generated/shap.plots.waterfall.html
        """
        if self.explainer.__class__.__name__ in ["Deep"]:
            shap_vals_as_exp = None
        else:
            shap_vals_as_exp = self.explainer(self.data)

        shap_values = self.shap_values
        if isinstance(shap_values, list) and len(shap_values) == 1:
            shap_values = shap_values[0]

        plt.close('all')

        if shap_vals_as_exp is None:

            features = self.features
            if not self.data_is_2d:
                features = self.unrolled_features

            class Explanation:
                # waterfall plot expects first argument as Explaination class
                # which must have at least these attributes (values, data, feature_names, base_values)
                # https://github.com/slundberg/shap/issues/1420#issuecomment-715190610
                if not self.data_is_2d:  # if original data is 3d then we flat it into 1d array
                    values = shap_values[example_index].reshape(-1, )
                    data = self.data[example_index].reshape(-1, )
                else:
                    values = shap_values[example_index]
                    data = self.data[example_index]

                feature_names = features
                base_values = self.explainer.expected_value[0]

            shap.plots.waterfall(Explanation(), show=False, max_display=max_display)
        else:
            shap.plots.waterfall(shap_vals_as_exp[example_index], show=False, max_display=max_display)

        if save:
            plt.savefig(os.path.join(self.path, f"{name}_{example_index}"), dpi=300, bbox_inches="tight")

        if show:
            plt.show()

        return

    def scatter_plot_single_feature(
            self, feature: str,
            name: str = "scatter",
            show=False,
            save=True,
            **scatter_kws
    ):

        shap_values = self.explainer(self.data)
        shap.plots.scatter(shap_values[:, feature], show=False, **scatter_kws)
        if save:
            plt.savefig(os.path.join(self.path, f"{name}_{feature}"), dpi=300, bbox_inches="tight")
        if show:
            plt.show()

        return

    def scatter_plot_all_features(self, name="scatter_plot", save=True, show=False, **scatter_kws):
        """draws scatter plot for all features"""
        if isinstance(self.data, pd.DataFrame):
            features = self.features
        else:
            features = [i for i in range(self.data.shape[-1])]

        for feature in features:
            self.scatter_plot_single_feature(feature, name=name, save=save, show=show, **scatter_kws)

        return

    def heatmap(self, name: str = 'heatmap', show: bool = False, save=False, max_display=10):
        """Plots the heatmap_ and saves it

        This can be drawn for xgboost/lgbm as well as for randomforest type models
        but not for CatBoostRegressor which is todo.

        Explanation
        ----------
        The upper line plot on the heat map shows $-fx/max(abs(fx))$ where $fx$ is
        the mean SHAP value of all features. The length of $fx$ is equal to length
        of data/examples. Thus one point on this line is the mean of SHAP values
        of all input features for the given/one example normalized by the maximum
        absolute value of $fx$.

        .. _heatmap:
            https://shap.readthedocs.io/en/latest/example_notebooks/api_examples/plots/heatmap.html
        """

        # if heat map is drawn with np.ndarray, it throws errors therefore convert
        # it into pandas DataFrame. It is more interpretable and does not hurt.
        shap_values = self._get_shap_values_locally()

        # by default examples are ordered in such a way that examples with similar
        # explanations are grouped together.
        self._heatmap(shap_values, f"{name}_basic",
                      show=show,
                      save=save,
                      max_display=max_display)

        # sort by the maximum absolute value of a feature over all the examples
        self._heatmap(shap_values, f"{name}_sortby_maxabs", show=show,
                      max_display=max_display,
                      save=save,
                      feature_values=shap_values.abs.max(0))

        # sorting by the sum of the SHAP values over all features gives a complementary perspective on the data
        self._heatmap(shap_values, f"{name}_sortby_SumOfShap", show=show,
                      save=save,
                      max_display=max_display,
                      instance_order=shap_values.sum(1))
        return

    def _heatmap(self, shap_values, name, show=False, max_display=10, save=True,  **kwargs):
        plt.close('all')

        # set show to False because we want to reset xlabel
        shap.plots.heatmap(shap_values, show=False,  max_display=max_display, **kwargs)
        plt.xlabel("Examples")

        if save:
            plt.savefig(os.path.join(self.path, f"{name}_sortby_SumOfShap"), dpi=300, bbox_inches="tight")

        if show:
            plt.show()
        return

    def _get_shap_values_locally(self):
        data = self.data
        if not isinstance(self.data, pd.DataFrame) and data.ndim == 2:
            data = pd.DataFrame(self.data, columns=self.features)

        # not using global explainer because, this explainer should data as well
        explainer = shap.Explainer(self.model, data)
        shap_values = explainer(data)

        return shap_values

    def beeswarm_plot(
            self,
            name: str = "beeswarm",
            show: bool = False,
            max_display: int = 10,
            save: bool = True,
            **kwargs
    ):
        """
        Draws the beeswarm_ plot of shap.

        Arguments:
            name : str
                name of saved file
            show : bool
                whether to show the plot or not
            save :
                whether to save the plot or not
            max_display :
                maximum
            kwargs :
                any keyword arguments for shap.beeswarm plot

        .. _beeswarm:
            https://shap.readthedocs.io/en/latest/example_notebooks/api_examples/plots/beeswarm.html)
        """

        shap_values = self._get_shap_values_locally()

        self._beeswarm_plot(shap_values,
                            name=f"{name}_basic", show=show, max_display=max_display,
                            save=save,
                            **kwargs)

        # find features with high impacts
        self._beeswarm_plot(shap_values, name=f"{name}_sortby_maxabs", show=show, max_display=max_display,
                            order=shap_values.abs.max(0), save=save, **kwargs)

        # plot the absolute value
        self._beeswarm_plot(shap_values.abs,
                            name=f"{name}_abs_shapvalues", show=show, max_display=max_display,
                            save=save, **kwargs)
        return

    def _beeswarm_plot(self, shap_values, name, show=False, max_display=10, save=True, **kwargs):

        plt.close('all')
        shap.plots.beeswarm(shap_values,  show=False, max_display=max_display, **kwargs)
        if save:
            plt.savefig(os.path.join(self.path, name), dpi=300, bbox_inches="tight")

        if show:
            plt.show()

        return

    def decision_plot(
            self,
            indices=None,
            name: str = "decision_",
            show=False,
            save=True,
            **decision_kwargs):
        """decision_ plot. For details see this blog.

        .. _decision:
            https://shap.readthedocs.io/en/latest/example_notebooks/api_examples/plots/decision_plot.html

        .. _blog:
            https://towardsdatascience.com/introducing-shap-decision-plots-52ed3b4a1cba
        """
        shap_values = self.shap_values
        legend_location = "best"
        legend_labels = None
        if indices is not None:
            shap_values = shap_values[(indices), :]
            if len(shap_values) <= 10:
                legend_labels = indices
                legend_location = "lower right"

        if self.explainer.__class__.__name__ in ["Tree"]:
            shap.decision_plot(self.explainer.expected_value,
                               shap_values,
                               self.features,
                               legend_labels=legend_labels,
                               show=False,
                               legend_location=legend_location,
                               **decision_kwargs)
            if save:
                plt.savefig(os.path.join(self.path, name), dpi=300, bbox_inches="tight")
            if show:
                plt.show()
        else:
            raise NotImplementedError

    def plot_shap_values(
            self,
            interpolation=None,
            cmap="coolwarm",
            name: str = "shap_values",
            show: bool = True,
            save: bool = False
    ):
        """Plots the SHAP values.

        Arguments:
            name:
                name of saved file
            show:
                whether to show the plot or not
            interpolation:
                interpolation argument to axis.imshow
            cmap:
                color map
            save:
                whether to save the plot or not

        """
        shap_values = self.shap_values

        if isinstance(shap_values, list) and len(shap_values) == 1:
            shap_values: np.ndarray = shap_values[0]

        def plot_shap_values_single_source(_data, _shap_vals, _features, _name):
            if _data.ndim == 3 and _shap_vals.ndim == 3:  # input is 3d
                # assert _shap_vals.ndim == 3
                return imshow_3d(_shap_vals,
                                 _data,
                                 _features,
                                 name=_name,
                                 path=self.path,
                                 show=show,
                                 cmap=cmap)

            plt.close('all')
            fig, axis = plt.subplots()
            im = axis.imshow(_shap_vals.T,
                             aspect='auto',
                             interpolation=interpolation,
                             cmap=cmap
                             )
            if _features is not None:  # if imshow is successful then don't worry if features are None
                axis.set_yticks(np.arange(len(_features)))
                axis.set_yticklabels(_features)
            axis.set_ylabel("Features")
            axis.set_xlabel("Examples")

            fig.colorbar(im)
            if save:
                plt.savefig(os.path.join(self.path, _name), dpi=300, bbox_inches="tight")

            if show:
                plt.show()
            return

        if self.single_source:
            plot_shap_values_single_source(self.data, shap_values, self.features, name)
        else:
            for idx, d in enumerate(self.data):
                plot_shap_values_single_source(d,
                                               shap_values[idx],
                                               self.features[idx],
                                               f"{idx}_{name}")
        return

    def pdp_all_features(
            self,
            show: bool = False,
            save: bool = True,
            **pdp_kws
    ):
        """partial dependence plot of all features.

        Arguments:
            show:
            save:
            pdp_kws:
                any keyword arguments
            """
        for feat in self.features:
            self.pdp_single_feature(feat, show=show, save=save, **pdp_kws)

        return

    def pdp_single_feature(
            self,
            feature_name: str,
            show=True,
            save=False,
            **pdp_kws
    ):
        """partial depence plot using SHAP package for a single feature."""

        shap_values = None
        if hasattr(self.shap_values, 'base_values'):
            shap_values = self.shap_values

        if self.model.__class__.__name__.startswith("XGB"):
            self.model.get_booster().feature_names = self.features

        fig = shap.partial_dependence_plot(
            feature_name,
            model=self.model.predict,
            data=self.data,
            model_expected_value=True,
            feature_expected_value=True,
            shap_values=shap_values,
            feature_names=self.features,
            show=False,
            **pdp_kws
        )

        if save:
            fname = f"pdp_{feature_name}"
            plt.savefig(os.path.join(self.path, fname), dpi=300, bbox_inches="tight")
        if show:
            plt.show()

        return fig


def imshow_3d(values,
              data,
              feature_names: list,
              path, vmin=None, vmax=None,
              name="",
              show=False,
              cmap=None,
              ):

    num_examples, lookback, input_features = values.shape
    assert data.shape == values.shape

    for idx, feat in enumerate(feature_names):
        plt.close('all')
        fig, (ax1, ax2) = plt.subplots(2, sharex='all', figsize=(10, 12))

        yticklabels=[f"t-{int(i)}" for i in np.linspace(lookback - 1, 0, lookback)]
        axis, im = imshow(data[:, :, idx].transpose(),
                          yticklabels=yticklabels,
                          ax=ax1,
                          vmin=vmin,
                          vmax=vmax,
                          title=feat,
                          cmap=cmap,
                          show=False
                          )
        fig.colorbar(im, ax=axis, orientation='vertical', pad=0.2)

        axis, im = imshow(values[:, :, idx].transpose(),
                          yticklabels=yticklabels,
                          vmin=vmin, vmax=vmax,
                          xlabel="Examples",
                          title=f"SHAP Values",
                          cmap=cmap,
                          show=False,
                          ax=ax2)

        fig.colorbar(im, ax=axis, orientation='vertical', pad=0.2)

        _name = f'{name}_{feat}_shap_values'
        plt.savefig(os.path.join(path, _name), dpi=400, bbox_inches='tight')

        if show:
            plt.show()
    return


def infer_framework(model):
    if hasattr(model, 'config') and 'backend' in model.config:
        framework = model.config['backend']

    elif type(model) is tuple:
        a, _ = model
        try:
            a.named_parameters()
            framework = 'pytorch'
        except:
            framework = 'tensorflow'
    else:
        try:
            model.named_parameters()
            framework = 'pytorch'
        except:
            framework = 'tensorflow'

    return framework


def maybe_to_dataframe(data, features=None) -> pd.DataFrame:
    if isinstance(data, np.ndarray) and isinstance(features, list) and data.ndim == 2:
        data = pd.DataFrame(data, columns=features)
    return data
